# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM Background command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


class BIM_Background:

    def GetResources(self):
        return {
            "Pixmap": "BIM_Background",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Background", "Toggle Background"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_Background",
                "Toggles the background of the 3D view between simple and gradient",
            ),
        }

    def Activated(self):
        param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View")
        if param.GetBool("Simple", True):
            param.SetBool("Simple", False)
            param.SetBool("Gradient", True)
            param.SetBool("RadialGradient", False)
        elif param.GetBool("Gradient", True):
            param.SetBool("Simple", False)
            param.SetBool("Gradient", False)
            param.SetBool("RadialGradient", True)
        else:
            param.SetBool("Simple", True)
            param.SetBool("Gradient", False)
            param.SetBool("RadialGradient", False)
        FreeCADGui.updateGui()


FreeCADGui.addCommand("BIM_Background", BIM_Background())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                 2020 Schildkroet                                        *
# *                 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import yaml
import pathlib
import FreeCAD
import FreeCADGui
import Path
from PySide.QtGui import (
    QStandardItem,
    QStandardItemModel,
    QPixmap,
    QDialog,
    QMessageBox,
    QWidget,
)
from PySide.QtCore import Qt, QEvent
from typing import List, cast, Tuple, Optional
from ...assets import AssetUri
from ...assets.ui import AssetOpenDialog, AssetSaveDialog
from ...camassets import cam_assets, ensure_assets_initialized
from ...shape.ui.shapeselector import ShapeSelector
from ...toolbit import ToolBit
from ...toolbit.serializers import all_serializers as toolbit_serializers
from ...toolbit.ui import ToolBitEditor
from ...toolbit.ui.toollist import ToolBitUriListMimeType
from ...toolbit.ui.util import natural_sort_key
from ..serializers import all_serializers as library_serializers
from ..models import Library
from .browser import LibraryBrowserWidget
from .properties import LibraryPropertyDialog


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


_LibraryRole = Qt.UserRole + 1
translate = FreeCAD.Qt.translate


class LibraryEditor(QWidget):
    """LibraryEditor is the controller for
    displaying/selecting/creating/editing a collection of ToolBits."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        Path.Log.track()
        ensure_assets_initialized(cam_assets)
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/ToolBitLibraryEdit.ui")
        self.form.installEventFilter(self)  # to forward keypress events
        self._base_title = self.form.windowTitle()

        # Create the library list.
        self.listModel = QStandardItemModel()
        self.form.TableList.setModel(self.listModel)
        self.form.TableList.clicked.connect(self._on_library_selected)

        # Enable drop support for the library list
        self.form.TableList.viewport().installEventFilter(self)  # Also on viewport

        # Create the LibraryBrowserWidget
        self.browser = LibraryBrowserWidget(
            asset_manager=cam_assets,
            parent=self,
        )
        self.browser.setDragEnabled(True)
        self.form.verticalLayout_2.layout().replaceWidget(self.form.toolTable, self.browser)
        self.form.toolTable.hide()

        # Connect signals.
        self.browser.itemDoubleClicked.connect(self.browser._on_edit_requested)

        self.form.addLibraryButton.clicked.connect(self._on_add_library_requested)
        self.form.removeLibraryButton.clicked.connect(self._on_remove_library_requested)
        self.form.renameLibraryButton.clicked.connect(self._on_rename_library_requested)
        self.form.importLibraryButton.clicked.connect(self._on_import_library_requested)
        self.form.exportLibraryButton.clicked.connect(self._on_export_library_requested)

        self.form.addToolBitButton.clicked.connect(self._on_add_toolbit_requested)
        self.form.importToolBitButton.clicked.connect(self._on_import_toolbit_requested)
        self.form.exportToolBitButton.clicked.connect(self._on_export_toolbit_requested)

        # Populate the UI.
        self._refresh_library_list()
        self._select_last_library()
        self._update_button_states()

    def _highlight_row(self, index):
        """Highlights the row at the given index using the selection model."""
        if not index.isValid():
            return
        self.form.TableList.setCurrentIndex(index)

    def _clear_highlight(self):
        """Clears the highlighting from the previously highlighted row."""
        self.form.TableList.selectionModel().clear()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress and self.form.TableList.hasFocus():
            if event.key() == Qt.Key_F2:
                Path.Log.debug("F2 pressed on library list.")
                self._on_rename_library_requested()
                return True
            elif event.key() == Qt.Key_Delete:
                Path.Log.debug("Del pressed on library list.")
                self._on_remove_library_requested()
                return True
        if obj == self.form.TableList.viewport():
            if event.type() == QEvent.DragEnter or event.type() == QEvent.DragMove:
                return self._handle_drag_enter(event)
            elif event.type() == QEvent.DragLeave:
                self._handle_drag_leave(event)
                return True
            elif event.type() == QEvent.Drop:
                return self._handle_drop(event)
        return super().eventFilter(obj, event)

    def _handle_drag_enter(self, event):
        """Handle drag enter and move events for the library list."""
        mime_data = event.mimeData()
        Path.Log.debug(f"_handle_drag_enter: MIME formats: {mime_data.formats()}")
        if not mime_data.hasFormat(ToolBitUriListMimeType):
            Path.Log.debug("_handle_drag_enter: Invalid MIME type, ignoring")
            return True

        # Get the row being hovered.
        pos = event.pos()
        event.acceptProposedAction()
        index = self.form.TableList.indexAt(pos)
        if not index.isValid():
            self._clear_highlight()
            return True

        # Prevent drop into "All Toolbits"
        item = self.listModel.itemFromIndex(index)
        if not item or item.data(_LibraryRole) == "all_tools":
            self._clear_highlight()
            return True

        self._highlight_row(index)
        return True

    def _handle_drag_leave(self, event):
        """Handle drag leave event for the library list."""
        self._clear_highlight()

    def _handle_drop(self, event):
        """Handle drop events to move or copy toolbits to the target library."""
        mime_data = event.mimeData()
        if not (mime_data.hasFormat(ToolBitUriListMimeType)):
            event.ignore()
            return True

        self._clear_highlight()
        pos = event.pos()
        index = self.form.TableList.indexAt(pos)
        if not index.isValid():
            event.ignore()
            return True

        item = self.listModel.itemFromIndex(index)
        if not item or item.data(_LibraryRole) == "all_tools":
            event.ignore()
            return True

        target_library_id = item.data(_LibraryRole)
        target_library_uri = f"toolbitlibrary://{target_library_id}"
        target_library = cast(Library, cam_assets.get(target_library_uri, depth=1))

        try:
            clipboard_content_yaml = mime_data.data(ToolBitUriListMimeType).data().decode("utf-8")
            clipboard_data_dict = yaml.safe_load(clipboard_content_yaml)

            if not isinstance(clipboard_data_dict, dict) or "toolbits" not in clipboard_data_dict:
                event.ignore()
                return True

            uris = clipboard_data_dict["toolbits"]
            new_uris = set()

            # Get the current library from the browser
            current_library = self.browser.get_current_library()

            for uri in uris:
                try:
                    toolbit = cast(ToolBit, cam_assets.get(AssetUri(uri), depth=0))
                    if toolbit:
                        added_toolbit = target_library.add_bit(toolbit)
                        if added_toolbit:
                            new_uris.add(str(toolbit.get_uri()))

                            # Remove the toolbit from the current library if it exists and
                            # it's not "all_tools"
                            if current_library and current_library.get_id() != "all_tools":
                                current_library.remove_bit(toolbit)
                except Exception as e:
                    Path.Log.error(f"Failed to load toolbit from URI {uri}: {e}")
                    continue

            if new_uris:
                cam_assets.add(target_library)
                # Save the current library if it was modified
                if current_library and current_library.get_id() != "all_tools":
                    cam_assets.add(current_library)
                self.browser.refresh()
                self.browser.select_by_uri(list(new_uris))
                self._update_button_states()

            event.acceptProposedAction()
        except Exception as e:
            Path.Log.error(f"Failed to process drop event: {e}")
            event.ignore()
        return True

    def get_selected_library_id(self) -> Optional[str]:
        index = self.form.TableList.currentIndex()
        if not index.isValid():
            return None
        item = self.listModel.itemFromIndex(index)
        if not item:
            return None
        return item.data(_LibraryRole)

    def get_selected_library(self, depth=1) -> Optional[Library]:
        library_id = self.get_selected_library_id()
        if not library_id:
            return None
        uri = f"toolbitlibrary://{library_id}"
        return cast(Library, cam_assets.get(uri, depth=depth))

    def select_library_by_uri(self, uri: AssetUri):
        # Find it in the list.
        index = 0
        for i in range(self.listModel.rowCount()):
            item = self.listModel.item(i)
            if item and item.data(_LibraryRole) == uri.asset_id:
                index = i
                break
        else:
            return

        # Select it.
        if index <= self.listModel.rowCount():
            item = self.listModel.item(index)
            if item:  # Should always be true, but...
                self.form.TableList.setCurrentIndex(self.listModel.index(index, 0))
                self._on_library_selected()

    def _select_last_library(self):
        # Find the last used library.
        last_used_lib_identifier = Path.Preferences.getLastToolLibrary()
        if last_used_lib_identifier:
            uri = Library.resolve_name(last_used_lib_identifier)
            self.select_library_by_uri(uri)

    def open(self):
        Path.Log.track()
        return self.form.exec_()

    def _refresh_library_list(self):
        """Clears and repopulates the self.listModel with available libraries."""
        Path.Log.track()
        self.listModel.clear()

        # Add "All Toolbits" item
        all_tools_item = QStandardItem(translate("CAM", "All Toolbits"))
        all_tools_item.setData("all_tools", _LibraryRole)
        # all_tools_item.setIcon(QPixmap(":/icons/CAM_ToolTable.svg"))
        # Make the "All Toolbits" item bold and italic
        font = all_tools_item.font()
        font.setBold(True)
        font.setItalic(True)
        all_tools_item.setFont(font)
        self.listModel.appendRow(all_tools_item)

        # Use AssetManager to fetch library assets (depth=0 for shallow fetch)
        try:
            # Fetch library assets themselves, not their deep dependencies (toolbits).
            # depth=0 means "fetch this asset, but not its dependencies"
            # The 'fetch' method returns actual Asset objects.
            libraries = cast(List[Library], cam_assets.fetch(asset_type="toolbitlibrary", depth=0))
        except Exception as e:
            Path.Log.error(f"Failed to fetch toolbit libraries: {e}")
            return

        # Sort by label for consistent ordering, falling back to asset_id if label is missing
        for library in sorted(
            libraries,
            key=lambda library: natural_sort_key(library.label or library.get_id()),
        ):
            lib_uri_str = str(library.get_uri())
            libItem = QStandardItem(library.label or library.get_id())
            libItem.setToolTip(f"ID: {library.get_id()}\nURI: {lib_uri_str}")
            libItem.setData(library.get_id(), _LibraryRole)  # Store the library ID
            libItem.setIcon(QPixmap(":/icons/CAM_ToolTable.svg"))
            self.listModel.appendRow(libItem)

        Path.Log.debug("model rows: {}".format(self.listModel.rowCount()))

        self.listModel.setHorizontalHeaderLabels(["Library"])

    def _on_library_selected(self):
        """Sets the current library in the browser when a library is selected."""
        Path.Log.debug("_on_library_selected: Called.")
        index = self.form.TableList.currentIndex()
        item = self.listModel.itemFromIndex(index)
        if not item:
            return
        if item.data(_LibraryRole) == "all_tools":
            selected_library = None
        else:
            selected_library = self.get_selected_library()
        self.browser.set_current_library(selected_library)
        self._update_window_title()
        self._update_button_states()

    def _update_window_title(self):
        """Updates the window title with the current library name."""
        current_library = self.browser.get_current_library()
        if current_library:
            title = f"{self._base_title} - {current_library.label}"
        else:
            title = self._base_title
        self.form.setWindowTitle(title)

    def _update_button_states(self):
        """Updates the enabled state of library management buttons."""
        library_selected = self.browser.get_current_library() is not None
        self.form.addLibraryButton.setEnabled(True)
        self.form.removeLibraryButton.setEnabled(library_selected)
        self.form.renameLibraryButton.setEnabled(library_selected)
        self.form.exportLibraryButton.setEnabled(library_selected)
        self.form.importLibraryButton.setEnabled(True)
        self.form.addToolBitButton.setEnabled(
            True
        )  # Always enabled - can create standalone toolbits
        # TODO: self.form.exportToolBitButton.setEnabled(toolbit_selected)

    def _save_library(self):
        """Internal method to save the current tool library asset"""
        Path.Log.track()
        library = self.browser.get_current_library()
        if not library:
            return

        # Save the modified library asset.
        try:
            cam_assets.add(library)
            Path.Log.debug(f"Library {library.get_uri()} saved")
        except Exception as e:
            Path.Log.error(f"Failed to save library {library.get_uri()}: {e}")
            QMessageBox.critical(
                self.form,
                translate("CAM_ToolBit", "Error Saving Library"),
                str(e),
            )
            raise

    def _on_add_library_requested(self):
        Path.Log.debug("_on_add_library_requested: Called.")
        new_library = Library(FreeCAD.Qt.translate("CAM", "New Library"))
        dialog = LibraryPropertyDialog(new_library, new=True, parent=self)
        if dialog.exec_() != QDialog.Accepted:
            return

        uri = cam_assets.add(new_library)
        Path.Log.debug(f"_on_add_library_requested: New library URI = {uri}")
        self._refresh_library_list()
        self.select_library_by_uri(uri)
        self._update_button_states()

    def _on_remove_library_requested(self):
        """Handles request to remove the selected library."""
        Path.Log.debug("_on_remove_library_requested: Called.")
        current_library = self.browser.get_current_library()
        if not current_library:
            return

        reply = QMessageBox.question(
            self,
            FreeCAD.Qt.translate("CAM", "Confirm Library Removal"),
            FreeCAD.Qt.translate(
                "CAM",
                "Are you sure you want to remove the library '{0}'?\n"
                "This will not delete the toolbits contained within it.",
            ).format(current_library.label),
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )

        if reply != QMessageBox.Yes:
            return

        try:
            library_uri = current_library.get_uri()
            cam_assets.delete(library_uri)
            Path.Log.info(f"Library {current_library.label} deleted.")
            self._refresh_library_list()
            self.browser.refresh()
            self._update_button_states()
        except FileNotFoundError as e:
            Path.Log.error(f"Failed to delete library {current_library.label}: {e}")
            QMessageBox.critical(
                self,
                FreeCAD.Qt.translate("CAM", "Error"),
                FreeCAD.Qt.translate("CAM", "Failed to delete library '{0}': {1}").format(
                    current_library.label, str(e)
                ),
            )

    def _on_rename_library_requested(self):
        """Handles request to rename the selected library."""
        Path.Log.debug("_on_rename_library_requested: Called.")
        current_library = self.browser.get_current_library()
        if not current_library:
            return

        dialog = LibraryPropertyDialog(current_library, new=False, parent=self)
        if dialog.exec_() != QDialog.Accepted:
            return

        cam_assets.add(current_library)
        self._refresh_library_list()
        self._update_button_states()

    def _on_import_library_requested(self):
        """Handles request to import a library."""
        Path.Log.debug("_on_import_library_requested: Called.")
        dialog = AssetOpenDialog(
            cam_assets, asset_class=Library, serializers=library_serializers, parent=self
        )
        response = dialog.exec_()
        if not response:
            return
        file_path, library = cast(Tuple[pathlib.Path, Library], response)

        try:
            cam_assets.add(library)
            self._refresh_library_list()
            self._update_button_states()
        except Exception as e:
            Path.Log.error(f"Failed to import library: {file_path} {e}")
            QMessageBox.critical(
                self,
                FreeCAD.Qt.translate("CAM", "Error"),
                FreeCAD.Qt.translate("CAM", f"Failed to import library: {file_path} {e}"),
            )

    def _on_export_library_requested(self):
        """Handles request to export the selected library."""
        Path.Log.debug("_on_export_library_requested: Called.")
        current_library = self.browser.get_current_library()
        if not current_library:
            return

        dialog = AssetSaveDialog(asset_class=Library, serializers=library_serializers, parent=self)
        dialog.exec_(current_library)
        self._update_button_states()

    def _on_add_toolbit_requested(self):
        """Handles request to add a new toolbit to the current library or create standalone."""
        Path.Log.debug("_on_add_toolbit_requested: Called.")
        current_library = self.browser.get_current_library()

        # Select the shape for the new toolbit
        selector = ShapeSelector()
        shape = selector.show()
        if shape is None:  # user canceled
            return

        try:
            # Find the appropriate ToolBit subclass based on the shape name
            tool_bit_classes = {b.SHAPE_CLASS.name: b for b in ToolBit.__subclasses__()}
            tool_bit_class = tool_bit_classes.get(shape.name)
            if not tool_bit_class:
                raise ValueError(f"No ToolBit subclass found for shape '{shape.name}'")

            # Create a new ToolBit instance
            new_toolbit = tool_bit_class(shape)
            new_toolbit.label = FreeCAD.Qt.translate("CAM", "New Toolbit")

            # Save the individual toolbit asset first
            tool_asset_uri = cam_assets.add(new_toolbit)
            Path.Log.debug(f"_on_add_toolbit_requested: Saved tool with URI: {tool_asset_uri}")

            # Add the toolbit to the current library if one is selected
            if current_library:
                toolno = current_library.add_bit(new_toolbit)
                Path.Log.debug(
                    f"_on_add_toolbit_requested: Added toolbit {new_toolbit.get_id()} (URI: {new_toolbit.get_uri()}) "
                    f"to current_library with number {toolno}."
                )
                # Save the library
                cam_assets.add(current_library)
            else:
                Path.Log.debug(
                    f"_on_add_toolbit_requested: Created standalone toolbit {new_toolbit.get_id()} (URI: {new_toolbit.get_uri()})"
                )

        except Exception as e:
            Path.Log.error(f"Failed to create or add new toolbit: {e}")
            QMessageBox.critical(
                self,
                FreeCAD.Qt.translate("CAM", "Error Creating Toolbit"),
                str(e),
            )
            raise

        self.browser.refresh()
        self.browser.select_by_uri([str(new_toolbit.get_uri())])
        self._update_button_states()

    def _on_import_toolbit_requested(self):
        """Handles request to import a toolbit."""
        Path.Log.debug("_on_import_toolbit_requested: Called.")
        current_library = self.browser.get_current_library()
        if not current_library:
            Path.Log.warning("Cannot import toolbit: No library selected.")
            QMessageBox.warning(
                self,
                FreeCAD.Qt.translate("CAM", "Warning"),
                FreeCAD.Qt.translate("CAM", "Please select a library first."),
            )
            return

        dialog = AssetOpenDialog(
            cam_assets, asset_class=ToolBit, serializers=toolbit_serializers, parent=self
        )
        response = dialog.exec_()
        if not response:
            return
        file_path, toolbit = cast(Tuple[pathlib.Path, ToolBit], response)

        # Debug logging for imported toolbit
        Path.Log.info(
            f"IMPORT TOOLBIT: file_path={file_path}, toolbit.id={toolbit.id}, toolbit.label={toolbit.label}"
        )
        import traceback

        stack = traceback.format_stack()
        caller_info = "".join(stack[-3:-1])
        Path.Log.info(f"IMPORT TOOLBIT CALLER:\n{caller_info}")

        # Check if toolbit already exists in asset manager
        toolbit_uri = toolbit.get_uri()
        Path.Log.info(f"IMPORT CHECK: toolbit_uri={toolbit_uri}")
        existing_toolbit = None
        try:
            existing_toolbit = cam_assets.get(toolbit_uri, store=["local", "builtin"], depth=0)
            Path.Log.info(
                f"IMPORT CHECK: Toolbit {toolbit.id} already exists, using existing reference"
            )
            Path.Log.info(
                f"IMPORT CHECK: existing_toolbit.id={existing_toolbit.id}, existing_toolbit.label={existing_toolbit.label}"
            )
        except FileNotFoundError:
            # Toolbit doesn't exist, save it as new
            Path.Log.info(f"IMPORT CHECK: Toolbit {toolbit.id} is new, saving to disk")
            new_uri = cam_assets.add(toolbit)
            Path.Log.info(f"IMPORT CHECK: Toolbit saved with new URI: {new_uri}")
            existing_toolbit = toolbit

        # Add the toolbit (existing or new) to the current library
        Path.Log.info(
            f"IMPORT ADD: Adding toolbit {existing_toolbit.id} to library {current_library.label}"
        )
        added_toolbit = current_library.add_bit(existing_toolbit)
        if added_toolbit:
            Path.Log.info(f"IMPORT ADD: Successfully added toolbit to library")
            cam_assets.add(current_library)  # Save the modified library
            self.browser.refresh()
            self.browser.select_by_uri([str(existing_toolbit.get_uri())])
            self._update_button_states()
        else:
            Path.Log.warning(f"IMPORT ADD: Failed to add toolbit {existing_toolbit.id} to library")
            Path.Log.warning(
                f"IMPORT FAILED: Failed to import toolbit from {file_path} to library {current_library.label}."
            )
            QMessageBox.warning(
                self,
                FreeCAD.Qt.translate("CAM", "Warning"),
                FreeCAD.Qt.translate(
                    "CAM",
                    f"Failed to import toolbit from '{file_path}' to library '{current_library.label}'.",
                ),
            )

    def _on_export_toolbit_requested(self):
        """Handles request to export the selected toolbit."""
        Path.Log.debug("_on_export_toolbit_requested: Called.")
        selected_toolbits = self.browser.get_selected_bits()
        if not selected_toolbits:
            Path.Log.warning("Cannot export toolbit: No toolbit selected.")
            QMessageBox.warning(
                self,
                FreeCAD.Qt.translate("CAM", "Warning"),
                FreeCAD.Qt.translate("CAM", "Please select a toolbit to export."),
            )
            return

        if len(selected_toolbits) > 1:
            Path.Log.warning("Cannot export multiple toolbits: Please select only one.")
            QMessageBox.warning(
                self,
                FreeCAD.Qt.translate("CAM", "Warning"),
                FreeCAD.Qt.translate("CAM", "Please select only one toolbit to export."),
            )
            return

        toolbit_to_export = selected_toolbits[0]
        dialog = AssetSaveDialog(asset_class=ToolBit, serializers=toolbit_serializers, parent=self)
        dialog.exec_(toolbit_to_export)  # This will open the save dialog and handle the export
        self._update_button_states()



# ***************************************************************************
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import ObjectsFem

from . import manager
from .ccx_cantilever_base_solid import setup_cantilever_base_solid
from .manager import init_doc


def get_information():
    return {
        "name": "CCX cantilever face load",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["fixed", "force"],
        "solvers": ["ccxtools", "elmer", "mystran", "z88"],
        "material": "solid",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.ccx_cantilever_faceload import setup
setup()


See forum topic post:
...

"""
    )


def setup(doc=None, solvertype="ccxtools", test_mode=False):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # setup CalculiX cantilever
    doc = setup_cantilever_base_solid(doc, solvertype, test_mode)
    analysis = doc.Analysis
    geom_obj = doc.Box

    # constraint force
    con_force = ObjectsFem.makeConstraintForce(doc, "ConstraintForce")
    con_force.References = [(geom_obj, "Face2")]
    con_force.Force = "9000000.0 N"
    con_force.Direction = (geom_obj, ["Edge5"])
    con_force.Reversed = True
    analysis.addObject(con_force)

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later
# /****************************************************************************
#                                                                           *
#    Copyright (c) 2025 Weston Schmidt <weston_schmidt@alumni.purdue.edu>   *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# ***************************************************************************/

"""
Unit tests for CommandInsertLink module.

This module contains unit tests to verify the proper handling of null
LinkedObject references in the TaskAssemblyInsertLink.accept() method.
Tests ensure that invalid objects are gracefully skipped without causing
AttributeError crashes.
"""

import unittest
from unittest.mock import patch, MagicMock

import FreeCAD as App

# Only import CommandInsertLink if GUI is available
if App.GuiUp:
    import CommandInsertLink


def _msg(text, end="\n"):
    """Write messages to the console including the line ending."""
    App.Console.PrintMessage(text + end)


@unittest.skipIf(not App.GuiUp, "GUI tests require FreeCAD GUI mode")
class TestCommandInsertLink(unittest.TestCase):
    """Unit tests for CommandInsertLink module."""

    @classmethod
    def setUpClass(cls):
        """setUpClass()...
        This method is called upon instantiation of this test class.  Add code and objects here
        that are needed for the duration of the test() methods in this class.  In other words,
        set up the 'global' test environment here; use the `setUp()` method to set up a 'local'
        test environment.
        This method does not have access to the class `self` reference, but it
        is able to call static methods within this same class.
        """

    @classmethod
    def tearDownClass(cls):
        """tearDownClass()...
        This method is called prior to destruction of this test class.  Add code and objects here
        that cleanup the test environment after the test() methods in this class have been executed.
        This method does not have access to the class `self` reference.  This method
        is able to call static methods within this same class.
        """

    def setUp(self):
        """setUp()...
        This method is called prior to each `test()` method.  Add code and objects here
        that are needed for multiple `test()` methods.
        """
        doc_name = self.__class__.__name__
        if App.ActiveDocument:
            if App.ActiveDocument.Name != doc_name:
                App.newDocument(doc_name)
        else:
            App.newDocument(doc_name)
        App.setActiveDocument(doc_name)
        self.doc = App.ActiveDocument

        self.assembly = App.ActiveDocument.addObject("Assembly::AssemblyObject", "Assembly")

        _msg(f"  Temporary document '{self.doc.Name}'")

    def tearDown(self):
        """tearDown()...
        This method is called after each test() method. Add cleanup instructions here.
        Such cleanup instructions will likely undo those in the setUp() method.
        """
        App.closeDocument(self.doc.Name)

    @patch("FreeCADGui.PySideUic.loadUi")
    @patch("CommandInsertLink.TaskAssemblyInsertLink.adjustTreeWidgetSize")
    def test_mixed_valid_and_invalid_objects(self, _mock_adjustTreeSize, _mock_loadUi):
        """Test that accept() handles a mix of valid and invalid objects correctly."""
        operation = "Handle mixed valid/invalid objects"
        _msg(f"  Test '{operation}'")

        mock_view = MagicMock()
        mock_view.getSize = MagicMock(return_value=(800, 600))
        task = CommandInsertLink.TaskAssemblyInsertLink(self.assembly, mock_view)

        # Create a mix of valid and invalid objects
        test_objects = [
            # Valid object (would work in real scenario)
            {
                "addedObject": type(
                    "ValidObject",
                    (),
                    {
                        "Name": "ValidObject",
                        "Label": "Valid Object",
                        "LinkedObject": type(
                            "ValidLinkedObject", (), {"Name": "ValidLinkedObjectName"}
                        )(),
                    },
                )(),
                "translation": App.Vector(1, 1, 1),
            },
            # Invalid: LinkedObject is None
            {
                "addedObject": type(
                    "InvalidObject1",
                    (),
                    {"Name": "InvalidObject1", "Label": "Invalid Object 1", "LinkedObject": None},
                )(),
                "translation": App.Vector(2, 2, 2),
            },
            # Invalid: No LinkedObject attribute
            {
                "addedObject": type(
                    "InvalidObject2", (), {"Name": "InvalidObject2", "Label": "Invalid Object 2"}
                )(),
                "translation": App.Vector(3, 3, 3),
            },
            # Invalid: LinkedObject.Name is None
            {
                "addedObject": type(
                    "InvalidObject3",
                    (),
                    {
                        "Name": "InvalidObject3",
                        "Label": "Invalid Object 3",
                        "LinkedObject": type("LinkedObjectWithNoneName", (), {"Name": None})(),
                    },
                )(),
                "translation": App.Vector(4, 4, 4),
            },
            # Invalid: No Name attribute
            {
                "addedObject": type(
                    "InvalidObject4",
                    (),
                    {
                        "Label": "Invalid Object 4",
                        "LinkedObject": type("LinkedObject", (), {"Name": "Something"})(),
                    },
                )(),
                "translation": App.Vector(5, 5, 5),
            },
        ]

        # Add all objects to insertion stack
        for obj_data in test_objects:
            task.insertionStack.append(obj_data)

        # Should handle the mix gracefully - invalid objects skipped, valid ones processed
        result = task.accept()
        self.assertTrue(result, "accept() should return True even with mixed valid/invalid objects")
        _msg("  Successfully handled mixed valid/invalid objects")

    @patch("FreeCADGui.PySideUic.loadUi")
    @patch("CommandInsertLink.TaskAssemblyInsertLink.adjustTreeWidgetSize")
    def test_empty_insertion_stack(self, _mock_adjustTreeSize, _mock_ui):
        """Test that accept() handles empty insertion stack correctly."""
        operation = "Handle empty insertion stack"
        _msg(f"  Test '{operation}'")

        mock_view = MagicMock()
        mock_view.getSize = MagicMock(return_value=(800, 600))
        task = CommandInsertLink.TaskAssemblyInsertLink(self.assembly, mock_view)

        # Don't add anything to insertion stack - it should remain empty
        self.assertEqual(len(task.insertionStack), 0, "Insertion stack should be empty")

        result = task.accept()
        self.assertTrue(result, "accept() should return True even with empty insertion stack")
        _msg("  Successfully handled empty insertion stack")



# SPDX-License-Identifier: LGPL-2.1-or-later

# Navigation indicator for FreeCAD
# Copyright (C) 2016, 2017, 2018 triplus @ FreeCAD
#
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Navigation indicator for FreeCAD."""

import Tux_rc
import FreeCAD as App
import FreeCADGui as Gui
from PySide import QtGui
from PySide import QtCore

mw = Gui.getMainWindow()
statusBar = mw.statusBar()
p = App.ParamGet("User parameter:Tux/NavigationIndicator")
pView = App.ParamGet("User parameter:BaseApp/Preferences/View")
pMWin = App.ParamGet("User parameter:BaseApp/Preferences/MainWindow")


def translate(context, text):
    "convenience function for Qt 5/6 translator"
    return QtGui.QApplication.translate(context, text, None)


class IndicatorButton(QtGui.QPushButton):
    """Detect language change events."""

    def __init__(self, parent=None):
        super(IndicatorButton, self).__init__()

    def changeEvent(self, event):
        """Change events."""
        if event.type() == QtCore.QEvent.LanguageChange:
            retranslateUi()
            onTooltip()
            self.adjustSize()
        return super(IndicatorButton, self).changeEvent(event)

    def onChange(self, paramGrp, param):
        if param == "NavigationStyle":
            setCurrent()

    def mousePressEvent(self, event):
        RePopulateIcons()
        return super(IndicatorButton, self).mousePressEvent(event)


def RePopulateIcons():
    curStyleSheet = pMWin.GetString("StyleSheet")
    if "dark" in curStyleSheet.lower():
        StyleSheetType = "light"
    else:
        StyleSheetType = "dark"

    a1.setIcon(QtGui.QIcon(":/icons/NavigationBlender_" + StyleSheetType + ".svg"))
    a2.setIcon(QtGui.QIcon(":/icons/NavigationCAD_" + StyleSheetType + ".svg"))
    a3.setIcon(QtGui.QIcon(":/icons/NavigationGesture_" + StyleSheetType + ".svg"))
    a4.setIcon(QtGui.QIcon(":/icons/NavigationMayaGesture_" + StyleSheetType + ".svg"))
    a5.setIcon(QtGui.QIcon(":/icons/NavigationOpenCascade_" + StyleSheetType + ".svg"))
    a6.setIcon(QtGui.QIcon(":/icons/NavigationOpenInventor_" + StyleSheetType + ".svg"))
    a7.setIcon(QtGui.QIcon(":/icons/NavigationOpenSCAD_" + StyleSheetType + ".svg"))
    a8.setIcon(QtGui.QIcon(":/icons/NavigationRevit_" + StyleSheetType + ".svg"))
    a9.setIcon(QtGui.QIcon(":/icons/NavigationSiemensNX_" + StyleSheetType + ".svg"))
    a10.setIcon(QtGui.QIcon(":/icons/NavigationSolidWorks_" + StyleSheetType + ".svg"))
    a11.setIcon(QtGui.QIcon(":/icons/NavigationTinkerCAD_" + StyleSheetType + ".svg"))
    a12.setIcon(QtGui.QIcon(":/icons/NavigationTouchpad_" + StyleSheetType + ".svg"))


def retranslateUi():
    """Retranslate navigation indicator on language change"""

    text01 = translate("NavigationIndicator", "Select")
    text02 = translate("NavigationIndicator", "Zoom")
    text03 = translate("NavigationIndicator", "Rotate")
    text04 = translate("NavigationIndicator", "Pan")
    text05 = translate("NavigationIndicator", "Tilt")
    text06 = translate("NavigationIndicator", "Navigation style")
    text07 = translate("NavigationIndicator", "Page Up or Page Down key.")
    text08 = translate("NavigationIndicator", "Rotation focus")
    text09 = translate("NavigationIndicator", "Middle mouse button or H key.")
    text10 = translate("NavigationIndicator", "Middle mouse button.")

    global t0
    t0 = translate("NavigationIndicator", "Navigation style not recognized.")

    global t1
    t1 = (
        "<p align='center'><b>Blender</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Middle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_ShiftMiddle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_LeftRight.svg'></td>
     </tr>
    </table>
    <b>"""
        + text08
        + ":</b> "
        + text10
        + "</small></p>"
    )

    global t2
    t2 = (
        "<p align='center'><b>CAD</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_MiddleLeft.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_MiddleRight.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Middle.svg'></td>
     </tr>
    </table>
    <b>"""
        + text08
        + ":</b> "
        + text10
        + "</small></p>"
    )

    global t3
    t3 = (
        "<p align='center'><b>Gesture</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
      <th><small>"""
        + text05
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_LeftMove.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_AltLeft.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Right.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_LeftRight.svg'></td>
     </tr>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
      <th><small>"""
        + text05
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Gesture_SelectTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_ZoomTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_RotateTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_PanTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_PanTouchAlt.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_TiltTouch.svg'></td>
     </tr>
    </table>
    <p><small><b>"""
        + text02
        + ":</b> "
        + text07
        + """<br>
    <b>"""
        + text08
        + ":</b> "
        + text09
        + "</small></p>"
    )

    global t4
    t4 = (
        "<p align='center'><b>MayaGesture</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
      <th><small>"""
        + text05
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_AltRight.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_AltLeft.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_AltMiddle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_AltLeftRight.svg'></td>
     </tr>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
      <th><small>"""
        + text05
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Gesture_SelectTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_ZoomTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_RotateTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_PanTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_PanTouchAlt.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Gesture_TiltTouch.svg'></td>
     </tr>
    </table>
    <p><small><b>"""
        + text02
        + ":</b> "
        + text07
        + """<br>
    <b>"""
        + text08
        + ":</b> "
        + text09
        + "</small></p>"
    )

    global t5
    t5 = (
        "<p align='center'><b>OpenCascade</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_CtrlLeft.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_CtrlRight.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_CtrlMiddle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Middle.svg'></td>
     </tr>
    </table>"""
    )

    global t6
    t6 = (
        "<p align='center'><b>OpenInventor</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_ShiftLeft.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_MiddleLeft.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Middle.svg'></td>
     </tr>
    </table>
    <b>"""
        + text08
        + ":</b> "
        + text10
        + "</small></p>"
    )

    global t7
    t7 = (
        "<p align='center'><b>OpenSCAD</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Middle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_MiddleRight.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Right.svg'></td>
     </tr>
    </table>"""
    )

    global t8
    t8 = (
        "<p align='center'><b>Revit</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_ShiftMiddle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Middle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_LeftRight.svg'></td>
     </tr>
    </table>
    <b>"""
        + text08
        + ":</b> "
        + text10
        + "</small></p>"
    )

    global t9
    t9 = (
        "<p align='center'><b>Siemens NX</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
        <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_MiddleLeft.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Middle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_MiddleRight.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_ShiftMiddle.svg'></td>
     </tr>
    </table>
    <b>"""
        + text08
        + ":</b> "
        + text10
        + "</small></p>"
    )

    global t10
    t10 = (
        "<p align='center'><b>SolidWorks</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_ShiftMiddle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Middle.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_CtrlMiddle.svg'></td>
     </tr>
    </table>
    <b>"""
        + text08
        + ":</b> "
        + text10
        + "</small></p>"
    )

    global t11
    t11 = (
        "<p align='center'><b>TinkerCAD</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Right.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Middle.svg'></td>
     </tr>
    </table>"""
    )

    global t12
    t12 = (
        "<p align='center'><b>Touchpad</b> "
        + text06
        + """</p>
    <table>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Mouse_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_Scroll.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_ShiftCtrlMove.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_AltMove.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_ShiftLeft.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Mouse_ShiftMove.svg'></td>
     </tr>
     <tr>
      <th><small>"""
        + text01
        + """</small></th>
      <th><small>"""
        + text02
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text03
        + """</small></th>
      <th><small>"""
        + text04
        + """</small></th>
     </tr>
     <tr>
      <td align='center'><img src=':/icons/Navigation_Touchpad_Left.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Touchpad_ShiftCtrlTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Touchpad_AltTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Touchpad_ShiftLeftTouch.svg'></td>
      <td align='center'><img src=':/icons/Navigation_Touchpad_ShiftTouch.svg'></td>
     </tr>
    </table>
    <p><small><b>"""
        + text02
        + ":</b> "
        + text07
        + "</p>"
    )

    menuSettings.setTitle(translate("NavigationIndicator", "Settings"))
    menuOrbit.setTitle(translate("NavigationIndicator", "Orbit style"))
    aCompact.setText(translate("NavigationIndicator", "Compact"))
    aTooltip.setText(translate("NavigationIndicator", "Tooltip"))
    aTurntable.setText(translate("NavigationIndicator", "Turntable"))
    aFreeTurntable.setText(translate("NavigationIndicator", "Free Turntable"))
    aTrackball.setText(translate("NavigationIndicator", "Trackball"))
    aTrackballClassic.setText(translate("NavigationIndicator", "Trackball Classic"))
    aRoundedArcball.setText(translate("NavigationIndicator", "Rounded Arcball"))
    a0.setText(translate("NavigationIndicator", "Undefined"))


indicator = IndicatorButton(statusBar)
indicator.setFlat(True)
indicator.adjustSize()
indicator.setObjectName("NavigationIndicator")
text = QtGui.QApplication.translate(
    "NavigationIndicator",
    "Navigation indicator",
    "A context menu action used to show or hide the 'Navigation indicator' toolbar widget",
)
indicator.setWindowTitle(text)

menu = QtGui.QMenu(indicator)
indicator.setMenu(menu)

menuSettings = QtGui.QMenu(menu)
menuOrbit = QtGui.QMenu(menu)

aCompact = QtGui.QAction(menuSettings)
aCompact.setCheckable(True)
aTooltip = QtGui.QAction(menuSettings)
aTooltip.setCheckable(True)

gOrbit = QtGui.QActionGroup(menuSettings)

aTurntable = QtGui.QAction(gOrbit)
aTurntable.setObjectName("NavigationIndicator_Turntable")
aTurntable.setCheckable(True)
aTrackball = QtGui.QAction(gOrbit)
aTrackball.setObjectName("NavigationIndicator_Trackball")
aTrackball.setCheckable(True)
aFreeTurntable = QtGui.QAction(gOrbit)
aFreeTurntable.setObjectName("NavigationIndicator_FreeTurntable")
aFreeTurntable.setCheckable(True)
aTrackballClassic = QtGui.QAction(gOrbit)
aTrackballClassic.setObjectName("NavigationIndicator_TrackballClassic")
aTrackballClassic.setCheckable(True)
aRoundedArcball = QtGui.QAction(gOrbit)
aRoundedArcball.setObjectName("NavigationIndicator_RoundedArcball")
aRoundedArcball.setCheckable(True)

menuOrbit.addAction(aRoundedArcball)
menuOrbit.addAction(aTrackball)
menuOrbit.addAction(aTrackballClassic)
menuOrbit.addAction(aTurntable)
menuOrbit.addAction(aFreeTurntable)

menuSettings.addMenu(menuOrbit)
menuSettings.addSeparator()
menuSettings.addAction(aCompact)
menuSettings.addAction(aTooltip)

gStyle = QtGui.QActionGroup(menu)

a0 = QtGui.QAction(gStyle)
a0.setIcon(QtGui.QIcon(":/icons/NavigationUndefined.svg"))
a0.setData("Undefined  ")
a0.setObjectName("Indicator_NavigationUndefined")

a1 = QtGui.QAction(gStyle)
a1.setText("Blender  ")
a1.setData("Gui::BlenderNavigationStyle")
a1.setObjectName("Indicator_NavigationBlender")

a2 = QtGui.QAction(gStyle)
a2.setText("CAD  ")
a2.setData("Gui::CADNavigationStyle")
a2.setObjectName("Indicator_NavigationCAD")

a3 = QtGui.QAction(gStyle)
a3.setText("Gesture  ")
a3.setData("Gui::GestureNavigationStyle")
a3.setObjectName("Indicator_NavigationGesture")

a4 = QtGui.QAction(gStyle)
a4.setText("MayaGesture  ")
a4.setData("Gui::MayaGestureNavigationStyle")
a4.setObjectName("Indicator_NavigationMayaGesture")

a5 = QtGui.QAction(gStyle)
a5.setText("OpenCascade  ")
a5.setData("Gui::OpenCascadeNavigationStyle")
a5.setObjectName("Indicator_NavigationOpenCascade")

a6 = QtGui.QAction(gStyle)
a6.setText("OpenInventor  ")
a6.setData("Gui::InventorNavigationStyle")
a6.setObjectName("Indicator_NavigationOpenInventor")

a7 = QtGui.QAction(gStyle)
a7.setText("OpenSCAD  ")
a7.setData("Gui::OpenSCADNavigationStyle")
a7.setObjectName("Indicator_NavigationOpenSCAD")

a8 = QtGui.QAction(gStyle)
a8.setText("Revit  ")
a8.setData("Gui::RevitNavigationStyle")
a8.setObjectName("Indicator_NavigationRevit")

a9 = QtGui.QAction(gStyle)
a9.setText("Siemens NX  ")
a9.setData("Gui::SiemensNXNavigationStyle")
a9.setObjectName("Indicator_NavigationSiemensNX")

a10 = QtGui.QAction(gStyle)
a10.setText("SolidWorks  ")
a10.setData("Gui::SolidWorksNavigationStyle")
a10.setObjectName("Indicator_NavigationSolidWorks")

a11 = QtGui.QAction(gStyle)
a11.setText("TinkerCAD  ")
a11.setData("Gui::TinkerCADNavigationStyle")
a11.setObjectName("Indicator_NavigationTinkerCAD")

a12 = QtGui.QAction(gStyle)
a12.setText("Touchpad  ")
a12.setData("Gui::TouchpadNavigationStyle")
a12.setObjectName("Indicator_NavigationTouchpad")

RePopulateIcons()

menu.addMenu(menuSettings)
menu.addSeparator()
menu.addAction(a0)
menu.addAction(a1)
menu.addAction(a2)
menu.addAction(a3)
menu.addAction(a4)
menu.addAction(a5)
menu.addAction(a6)
menu.addAction(a7)
menu.addAction(a8)
menu.addAction(a9)
menu.addAction(a10)
menu.addAction(a11)
menu.addAction(a12)

pView.Attach(indicator)


def onCompact():
    """Enable or disable compact mode."""

    if aCompact.isChecked():
        p.SetBool("Compact", 1)
    else:
        p.SetBool("Compact", 0)

    setCurrent()


def setCompact(action):
    """Set compact mode."""

    if p.GetBool("Compact", 0):
        indicator.setText("")
    else:
        indicator.setText(action.text())
        indicator.adjustSize()


def onTooltip():
    """Enable or disable verbose tooltips."""

    if aTooltip.isChecked():
        a0.setToolTip(t0)
        a1.setToolTip(t1)
        a2.setToolTip(t2)
        a3.setToolTip(t3)
        a4.setToolTip(t4)
        a5.setToolTip(t5)
        a6.setToolTip(t6)
        a7.setToolTip(t7)
        a8.setToolTip(t8)
        a9.setToolTip(t9)
        a10.setToolTip(t10)
        a11.setToolTip(t11)
        a12.setToolTip(t12)
        p.SetBool("Tooltip", 1)
    else:
        for i in gStyle.actions():
            i.setToolTip("")
        p.SetBool("Tooltip", 0)

    setCurrent()


def onOrbit():
    """Use turntable or trackball orbit style."""

    if aTurntable.isChecked():
        pView.SetInt("OrbitStyle", 0)
    elif aTrackball.isChecked():
        pView.SetInt("OrbitStyle", 1)
    elif aFreeTurntable.isChecked():
        pView.SetInt("OrbitStyle", 2)
    elif aTrackballClassic.isChecked():
        pView.SetInt("OrbitStyle", 3)
    elif aRoundedArcball.isChecked():
        pView.SetInt("OrbitStyle", 4)


def onOrbitShow():
    """Set turntable or trackball orbit style."""

    OrbitStyle = pView.GetInt("OrbitStyle", 4)
    gOrbit.blockSignals(True)
    if OrbitStyle == 0:
        aTurntable.setChecked(True)
    elif OrbitStyle == 1:
        aTrackball.setChecked(True)
    elif OrbitStyle == 2:
        aFreeTurntable.setChecked(True)
    elif OrbitStyle == 3:
        aTrackballClassic.setChecked(True)
    elif OrbitStyle == 4:
        aRoundedArcball.setChecked(True)
    gOrbit.blockSignals(False)


def onMenu(action):
    """Set navigation style on selection."""
    pView.SetString("NavigationStyle", action.data())


def setCurrent():
    """Set navigation style on start and on interval."""
    gStyle.blockSignals(True)

    s = False
    actions = gStyle.actions()
    current = pView.GetString("NavigationStyle")

    if current and current != "Undefined":
        for i in actions:
            if i.data() == current:
                s = True
                setCompact(i)
                menu.setDefaultAction(i)
                indicator.setIcon(i.icon())
                indicator.setToolTip(i.toolTip())
            else:
                pass
    else:
        s = True
        pView.SetString("NavigationStyle", a2.data())

    if s:
        a0.setVisible(False)
    else:
        a0.setVisible(True)
        a0.setEnabled(True)
        setCompact(a0)
        menu.setDefaultAction(a0)
        indicator.setIcon(a0.icon())
        indicator.setToolTip(a0.toolTip())

    gStyle.blockSignals(False)


if p.GetBool("Compact", 0):
    aCompact.setChecked(True)

if p.GetBool("Tooltip", 1):
    aTooltip.setChecked(True)

retranslateUi()
onCompact()
onTooltip()

label = statusBar.children()[2]
statusBar.removeWidget(label)
statusBar.addPermanentWidget(indicator)
statusBar.addPermanentWidget(label)
label.show()

setCurrent()

gStyle.triggered.connect(onMenu)
gOrbit.triggered.connect(onOrbit)
aCompact.triggered.connect(onCompact)
aTooltip.triggered.connect(onTooltip)
menuOrbit.aboutToShow.connect(onOrbitShow)
menu.aboutToHide.connect(indicator.clearFocus)



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Provides classes and support functions for managing the automatically installed
Python library dependencies. No support is provided for uninstalling those dependencies
because pip's uninstall function does not support the target directory argument."""

import dataclasses
import os
import re
import shutil
import subprocess
from typing import Dict, Iterable, List, TypedDict, Optional, Set
from addonmanager_metadata import Version
from addonmanager_utilities import (
    create_pip_call,
    run_interruptable_subprocess,
    get_pip_target_directory,
    translate,
    using_system_pip_installation_location,
)

import addonmanager_freecad_interface as fci

from PySideWrapper import QtCore

translate = fci.translate


class PipFailed(Exception):
    """Exception thrown when pip times out or otherwise fails to return valid results"""


def pip_has_dry_run_support() -> bool:
    """Returns True if pip supports the --dry-run option, False otherwise."""
    try:
        pip_version_string = call_pip(["--version"])[0]
        version_str = pip_version_string.split()[1]
        return Version(version_str) >= Version("23.1")
    except PipFailed:
        return False


def call_pip(args: List[str]) -> List[str]:
    """Tries to locate the appropriate Python executable and run pip with version checking
    disabled. Fails if Python can't be found or if pip is not installed."""

    try:
        call_args = create_pip_call(args)
        fci.Console.PrintLog(f"Running pip with the following command:\n")
        fci.Console.PrintLog(" ".join(call_args) + "\n")
    except RuntimeError as exception:
        raise PipFailed() from exception

    try:
        proc = run_interruptable_subprocess(call_args, 120)
    except subprocess.CalledProcessError as exception:
        raise PipFailed(f"pip call failed:\n{exception}") from exception

    if proc.returncode != 0:
        raise PipFailed(proc.stderr)

    data = proc.stdout
    return data.split("\n")


@dataclasses.dataclass
class PackageInfo:
    name: str
    installed_version: str
    available_version: str
    dependencies: List[str]


def pep503_normalize(package_name: str) -> str:
    """Given a Python package name, normalize it per PEP 503, making it all lowercase, and replacing
    underscores and dots with dashes."""

    result = package_name.replace("_", "-")
    result = result.replace(".", "-")
    return result.lower()


def parse_pip_list_output(all_packages, outdated_packages) -> List[PackageInfo]:
    """Parses the output from pip into a dictionary with update information in it. The pip
    output should be an array of lines of text."""

    # All Packages output looks like this:
    # Package    Version
    # ---------- -------
    # gitdb      4.0.9
    # setuptools 41.2.0

    # Outdated Packages output looks like this:
    # Package    Version Latest Type
    # ---------- ------- ------ -----
    # pip        21.0.1  22.1.2 wheel
    # setuptools 41.2.0  63.2.0 wheel

    packages: Dict[str, PackageInfo] = {}
    skip_counter = 0
    for line in all_packages:
        if skip_counter < 2:
            skip_counter += 1
            continue
        entries = line.split()
        if len(entries) > 1:
            package_name = pep503_normalize(entries[0])
            installed_version = entries[1]
            packages[package_name] = PackageInfo(package_name, installed_version, "", [])

    skip_counter = 0
    for line in outdated_packages:
        if skip_counter < 2:
            skip_counter += 1
            continue
        entries = line.split()
        if len(entries) > 1:
            package_name = pep503_normalize(entries[0])
            available_version = entries[2]
            if package_name not in packages:
                raise RuntimeError(
                    "all_packages does not contain all packages in outdated_packages"
                )
            packages[package_name].available_version = available_version

    return list(packages.values())


class AsynchronousResetWorker(QtCore.QObject):
    """A worker class that runs pip to generate the package list."""

    finished = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.is_running = False
        self.error = ""
        self.vendor_path = get_pip_target_directory()
        self.package_list = []

    def run(self):
        """Runs pip: when complete, either self.package_list is populated, or self.error is set."""
        self.is_running = True
        self.error = ""
        self.package_list = []
        try:
            outdated_packages_stdout = call_pip(["list", "-o", "--path", self.vendor_path])
            all_packages_stdout = call_pip(["list", "--path", self.vendor_path])
            self.package_list = parse_pip_list_output(all_packages_stdout, outdated_packages_stdout)
        except PipFailed as e:
            self.error = str(e)
        self.is_running = False
        self.finished.emit()


class AsynchronousUpdateWorker(QtCore.QObject):
    """A worker class that runs pip to generate the package list."""

    finished = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.is_running = False
        self.error = ""
        self.vendor_path = get_pip_target_directory()
        self.package_list = []

    def run(self):
        """Runs pip: when complete, either self.package_list is populated, or self.error is set."""
        self.is_running = True
        self.error = ""

        update_string = " ".join(self.package_list)
        try:
            fci.Console.PrintLog(
                f"Running pip to upgrade the following packages in {self.vendor_path}: {update_string}\n"
            )
            command = ["install", "--upgrade", "--target", self.vendor_path]
            command.extend(self.package_list)
            upgrade_stdout = call_pip(command)
            for line in upgrade_stdout:
                fci.Console.PrintLog(line + "\n")
        except PipFailed as e:
            self.error = str(e)
            fci.Console.PrintError(self.error + "\n")

        try:
            outdated_packages_stdout = call_pip(["list", "-o", "--path", self.vendor_path])
            all_packages_stdout = call_pip(["list", "--path", self.vendor_path])
            self.package_list = parse_pip_list_output(all_packages_stdout, outdated_packages_stdout)
        except PipFailed as e:
            self.error = str(e)
        self.is_running = False
        self.finished.emit()


class PythonPackageListModel(QtCore.QAbstractTableModel):
    """The non-GUI portion of the Python package manager. This class is responsible for
    communicating with pip and generating a list of packages to be installed, acting as a model
    for the Qt view."""

    update_complete = QtCore.Signal()

    def __init__(self, addons):
        super().__init__()
        self.addons = addons
        self.is_venv = False
        self.vendor_path = get_pip_target_directory()  # Ignored if running in a venv
        self.package_list = []
        self.reset_worker = None
        self.update_worker = None
        self.reset_worker_thread = None
        self.update_worker_thread = None

    def can_use_thread(self) -> bool:
        threaded = (
            QtCore.QCoreApplication.instance() is not None
            and QtCore.QCoreApplication.instance().thread().isRunning()
        )
        return threaded

    def reset_package_list(self):
        """Reset the model: asynchronous if the GUI is running (that is, if QThreads can be used),
        otherwise synchronous."""
        self.beginResetModel()
        self.package_list.clear()
        self.reset_worker = AsynchronousResetWorker()
        if self.can_use_thread():
            self.reset_worker_thread = QtCore.QThread()
            self.reset_worker.moveToThread(self.reset_worker_thread)
            self.reset_worker_thread.started.connect(self.reset_worker.run)
            self.reset_worker.finished.connect(self.reset_call_finished)
            self.reset_worker.finished.connect(self.reset_worker_thread.quit)
            self.reset_worker_thread.start()
        else:
            self.reset_worker.run()
            self.reset_call_finished()

    def reset_call_finished(self):
        if self.reset_worker.error:
            fci.Console.PrintError(f"Error while resetting package list: {self.reset_worker.error}")
        self.package_list = self.reset_worker.package_list
        self.endResetModel()

    def rowCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        if parent.isValid():
            return 0
        return len(self.package_list)

    def columnCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        if parent.isValid():
            return 0
        return 4

    def data(self, index, role=...) -> Optional[str]:
        row = index.row()
        col = index.column()
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if col == 0:
                return self.package_list[row].name
            elif col == 1:
                return self.package_list[row].installed_version
            elif col == 2:
                return self.package_list[row].available_version
            elif col == 3:
                if not self.package_list[row].dependencies:
                    dependent_addons = self.get_dependent_addons(self.package_list[row].name)
                    for addon in dependent_addons:
                        if addon["optional"]:
                            self.package_list[row].dependencies.append(addon["name"] + "*")
                        else:
                            self.package_list[row].dependencies.append(addon["name"])
                return ", ".join(self.package_list[row].dependencies)
        return None

    def headerData(self, section, orientation, role=...) -> Optional[str]:
        if (
            orientation == QtCore.Qt.Orientation.Horizontal
            and role == QtCore.Qt.ItemDataRole.DisplayRole
        ):
            if section == 0:
                return translate("AddonsInstaller", "Package")
            elif section == 1:
                return translate("AddonsInstaller", "Installed Version")
            elif section == 2:
                return translate("AddonsInstaller", "Available Version")
            elif section == 3:
                return translate("AddonsInstaller", "Dependencies")
        return None

    def flags(self, index) -> QtCore.Qt.ItemFlag:
        return QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable

    def updates_are_available(self) -> bool:
        """Returns True if there are updates available for any packages, False otherwise."""
        for package in self.package_list:
            if package.available_version:
                return True
        return False

    class DependentAddon(TypedDict):
        name: str
        optional: bool

    def get_dependent_addons(self, package) -> List[DependentAddon]:
        dependent_addons = []
        for addon in self.addons:
            # if addon.installed_version is not None:
            if package in [pep503_normalize(x) for x in addon.python_requires]:
                dependent_addons.append({"name": addon.name, "optional": False})
            elif package in [pep503_normalize(x) for x in addon.python_optional]:
                dependent_addons.append({"name": addon.name, "optional": True})
        return dependent_addons

    def update_all_packages(self):
        """Re-installs all packages. Uses an asynchronous thread when possible."""
        self.update_worker = AsynchronousUpdateWorker()

        updates = [item.name for item in self.package_list]
        self.update_worker.package_list = updates
        if not using_system_pip_installation_location():
            # pip doesn't properly update when using the target directory, so we have to delete
            # it and reinstall
            os.rename(self.vendor_path, self.vendor_path + ".old")
            os.mkdir(self.vendor_path)
        if self.can_use_thread():
            self.update_worker_thread = QtCore.QThread()
            self.update_worker.moveToThread(self.update_worker_thread)
            self.update_worker_thread.started.connect(self.update_worker.run)
            self.update_worker.finished.connect(self.update_call_finished)
            self.update_worker.finished.connect(self.update_worker_thread.quit)
            self.update_worker_thread.start()
        else:
            self.update_worker.run()
            self.update_call_finished()

    def update_call_finished(self):
        self.update_complete.emit()
        if not using_system_pip_installation_location():
            shutil.rmtree(self.vendor_path + ".old")

    def determine_new_python_dependencies(self, addons) -> Set[str]:
        """Given a list of Addon objects, finds the Python dependencies for those addons. Also
        accepts a single Addon object, in which case only its dependencies are evaluated. If using
        a recent version of pip, the dry-run option is used to determine which packages would be
        installed, otherwise just lists the dependencies as they are listed in the
        addon metadata, filtered to only show new ones."""

        if not isinstance(addons, Iterable):
            addons = [addons]

        python_dependencies = set()
        for addon in addons:
            python_dependencies.update(addon.python_requires)
            python_dependencies.update(addon.python_optional)

        result = set()
        # If we have at least pip 23.1, we can use dry-run:
        if pip_has_dry_run_support():
            command = ["install", "--upgrade", "--dry-run"]
            command.extend(python_dependencies)
            output = call_pip(command)

            for line in output:
                match = re.match(r"Would install ([\w\-.]+)-([\d.]+)", line)
                if match:
                    name, _ = match.groups()
                    result.add(name)
        else:
            result.update(python_dependencies)
            result.difference_update(set([p.name for p in self.package_list]))
        return result

    def all_dependencies_installed(self, addon) -> bool:
        """Returns True if all dependencies for the given addon are installed, or False if not."""
        dependencies = self.determine_new_python_dependencies(addon)
        return len(dependencies) == 0



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 David Carter <dcarter@davidcarter.ca>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__author__ = "David Carter"
__url__ = "https://www.davesrocketshop.com"

from abc import ABC, abstractmethod
from dataclasses import dataclass

import Materials

@dataclass
class MaterialLibraryType:
    name: str
    icon: bytes
    readOnly: bool

@dataclass
class MaterialLibraryObjectType:
    UUID: str
    path: str
    name: str

@dataclass
class ModelObjectType:
    libraryName: str
    model: Materials.Model

@dataclass
class MaterialObjectType:
    libraryName: str
    material: Materials.Material

class MaterialManagerExternal(ABC):
    """Abstract base class for all external material managers

    Any external interface should be derivedfrom this base class."""

    @classmethod
    def APIVersion(cls) -> tuple:
        """Returns a tuple of 3 integers describing the API version

        The version returned should be the latest supported version. This method
        allows the interface to use older modules."""
        return (1, 0, 0)

    #
    # Library methods
    #

    @abstractmethod
    def libraries(self) -> list[MaterialLibraryType]:
        """Returns a list of libraries managed by this interface

        The list contains a series of tuples describing all libraries managed by
        this module. Each tuple contains the library name, icon, a boolean to indicate
        if it is a read only library."""

    @abstractmethod
    def modelLibraries(self) -> list[MaterialLibraryType]:
        """Returns a list of libraries managed by this interface

        The list contains a series of tuples describing all libraries managed by
        this module. Each tuple contains the library name, icon, and a boolean to indicate
        if it is a read only library.

        This differs from the libraries() function in that it only returns libraries
        containing model objects."""

    @abstractmethod
    def materialLibraries(self) -> list[MaterialLibraryType]:
        """Returns a list of libraries managed by this interface

        The list contains a series of tuples describing all libraries managed by
        this module. Each tuple contains the library name, icon, and a boolean to indicate
        if it is a read only library.

        This differs from the libraries() function in that it only returns libraries
        containing material objects."""

    @abstractmethod
    def getLibrary(self, libraryName: str) -> MaterialLibraryType:
        """Get the library

        Retrieve the library with the given name"""

    @abstractmethod
    def createLibrary(self, libraryName: str, icon: bytes, readOnly: bool) -> None:
        """Create a new library

        Create a new library with the given name."""

    @abstractmethod
    def renameLibrary(self, oldName: str, newName: str) -> None:
        """Rename an existing library

        Change the name of an existing library"""

    @abstractmethod
    def changeIcon(self, libraryName: str, icon: bytes) -> None:
        """Change the library icon

        Change the library icon"""

    @abstractmethod
    def removeLibrary(self, libraryName: str) -> None:
        """Delete a library and its contents

        Deletes the library and any models or materials it contains"""

    @abstractmethod
    def libraryModels(self, libraryName: str) -> list[MaterialLibraryObjectType]:
        """Returns a list of models managed by this library

        Each list entry is a tuple containing the UUID, path, and name of the model"""

    @abstractmethod
    def libraryMaterials(self, libraryName: str,
                         materialFilter: Materials.MaterialFilter = None,
                         options: Materials.MaterialFilterOptions = None) -> \
                                list[MaterialLibraryObjectType]:
        """Returns a list of materials managed by this library

        Each list entry is a tuple containing the UUID, path, and name of the material"""

    @abstractmethod
    def libraryFolders(self, libraryName: str) -> list[str]:
        """Returns a list of folders managed by this library

        This will return a list of all folders in the library including empty folders"""

    #
    # Folder methods
    #

    @abstractmethod
    def createFolder(self, libraryName: str, path: str) -> None:
        """Create a new folder in the given library"""

    @abstractmethod
    def renameFolder(self, libraryName: str, oldPath: str, newPath: str) -> None:
        """Rename the folder"""

    @abstractmethod
    def deleteRecursive(self, libraryName: str, path: str) -> None:
        """Delete the folder and all of its contents"""

    #
    # Model methods
    #

    @abstractmethod
    def getModel(self, uuid: str) -> ModelObjectType:
        """Retrieve a model given its UUID"""

    @abstractmethod
    def addModel(self, libraryName: str, path: str, model: Materials.Model) -> None:
        """Add a model to a library in the given folder. The folder path is relative to
        the library and will be created if it doesn't already exist.

        This will throw a DatabaseModelExistsError exception if the model already exists."""

    @abstractmethod
    def migrateModel(self, libraryName: str, path: str, model: Materials.Model) -> None:
        """Add the model to the library.

        If the model already exists, then no action is performed."""

    @abstractmethod
    def updateModel(self, libraryName: str, path: str, model: Materials.Model) -> None:
        """Update the given model"""

    @abstractmethod
    def setModelPath(self, libraryName: str, path: str, uuid: str) -> None:
        """Change the model path within the library"""

    @abstractmethod
    def renameModel(self, libraryName: str, name: str, uuid: str) -> None:
        """Change the model name"""

    @abstractmethod
    def moveModel(self, libraryName: str, path: str, uuid: str) -> None:
        """Move a model across libraries

        Move the model to the desired path in a different library. This should also
        remove the model from the old library if that library is managed by this
        interface"""

    @abstractmethod
    def removeModel(self, uuid: str) -> None:
        """Remove the model from the library"""

    #
    # Material methods
    #

    @abstractmethod
    def getMaterial(self, uuid: str) -> MaterialObjectType:
        """ Retrieve a material given its UUID """

    @abstractmethod
    def addMaterial(self, libraryName: str, path: str, material: Materials.Material) -> None:
        """Add a material to a library in the given folder. The folder path is relative to
        the library and will be created if it doesn't already exist.

        This will throw a DatabaseMaterialExistsError exception if the model already exists."""

    @abstractmethod
    def migrateMaterial(self, libraryName: str, path: str, material: Materials.Material) -> None:
        """Add the material to the library in the given folder. The folder path is relative to
        the library and will be created if it doesn't already exist.

        If the material already exists, then no action is performed."""

    @abstractmethod
    def updateMaterial(self, libraryName: str, path: str, material: Materials.Material) -> None:
        """Update the given material"""

    @abstractmethod
    def setMaterialPath(self, libraryName: str, path: str, uuid: str) -> None:
        """Change the material path within the library"""

    @abstractmethod
    def renameMaterial(self, libraryName: str, name: str, uuid: str) -> None:
        """Change the material name"""

    @abstractmethod
    def moveMaterial(self, libraryName: str, path: str, uuid: str) -> None:
        """Move a material across libraries

        Move the material to the desired path in a different library. This should also
        remove the material from the old library if that library is managed by this
        interface"""

    @abstractmethod
    def removeMaterial(self, uuid: str) -> None:
        """Remove the material from the library"""



# ***************************************************************************
# *   Copyright (c) 2020 Adam Spontarelli <adam@vector-space.org>           *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD, Part
from fcsprocket import fcsprocket
from fcsprocket import sprocket

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtCore, QtGui
    from FreeCADGui import PySideUic as uic

__title__ = "PartDesign SprocketObject management"
__author__ = "Adam Spontarelli"
__url__ = "https://www.freecad.org"


def makeSprocket(name):
    """
    makeSprocket(name): makes a Sprocket
    """
    obj = FreeCAD.ActiveDocument.addObject("Part::Part2DObjectPython", name)
    Sprocket(obj)
    if FreeCAD.GuiUp:
        ViewProviderSprocket(obj.ViewObject)
    # FreeCAD.ActiveDocument.recompute()
    if FreeCAD.GuiUp:
        body = FreeCADGui.ActiveDocument.ActiveView.getActiveObject("pdbody")
        part = FreeCADGui.ActiveDocument.ActiveView.getActiveObject("part")
        if body:
            body.Group = body.Group + [obj]
        elif part:
            part.Group = part.Group + [obj]
    return obj


class CommandSprocket:
    """
    the Fem Sprocket command definition
    """

    def GetResources(self):
        return {
            "Pixmap": "PartDesign_Sprocket",
            "MenuText": QtCore.QT_TRANSLATE_NOOP("PartDesign_Sprocket", "Sprocket"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "PartDesign_Sprocket", "Creates or edits the sprocket definition."
            ),
        }

    def Activated(self):

        FreeCAD.ActiveDocument.openTransaction("Create Sprocket")
        FreeCADGui.addModule("SprocketFeature")
        FreeCADGui.doCommand("SprocketFeature.makeSprocket('Sprocket')")
        FreeCADGui.doCommand("Gui.activeDocument().setEdit(App.ActiveDocument.ActiveObject.Name,0)")

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


class Sprocket:
    """
    The Sprocket object
    """

    """
        ANSI B29.1-2011 standard roller chain sizes in USCS units (inches)
        {size: [Pitch, Roller Diameter]}
        """
    SprocketReferenceRollerTable = {
        0: [0.250, 0.130, 0.110, "ANSI 25"],
        1: [0.375, 0.200, 0.168, "ANSI 35"],
        2: [0.500, 0.306, 0.227, "ANSI 41"],
        3: [0.500, 0.312, 0.284, "ANSI 40"],
        4: [0.625, 0.400, 0.343, "ANSI 50"],
        5: [0.750, 0.469, 0.459, "ANSI 60"],
        6: [1.000, 0.625, 0.575, "ANSI 80"],
        7: [1.250, 0.750, 0.692, "ANSI 100"],
        8: [1.500, 0.875, 0.924, "ANSI 120"],
        9: [1.750, 1.000, 0.924, "ANSI 140"],
        10: [2.000, 1.125, 1.156, "ANSI 160"],
        11: [2.250, 1.460, 1.301, "ANSI 180"],
        12: [2.500, 1.562, 1.389, "ANSI 200"],
        13: [3.000, 1.875, 1.738, "ANSI 240"],
        14: [0.500, 0.3125, 0.11, "Bicycle with Derailleur"],
        15: [0.500, 0.3125, 0.084, "Bicycle without Derailleur"],
        16: [0.375, 5.72 / 25.4, 5.2 / 25.4, "ISO 606 06B"],
        17: [0.500, 7.75 / 25.4, 7.0 / 25.4, "ISO 606 08B"],
        18: [0.625, 9.65 / 25.4, 9.1 / 25.4, "ISO 606 10B"],
        19: [0.750, 11.68 / 25.4, 11.1 / 25.4, "ISO 606 12B"],
        20: [1.000, 17.02 / 25.4, 16.2 / 25.4, "ISO 606 16B"],
        21: [1.250, 19.56 / 25.4, 18.5 / 25.4, "ISO 606 20B"],
        22: [1.500, 25.4 / 25.4, 24.1 / 25.4, "ISO 606 24B"],
        23: [0.500, 0.3125, 0.227, "Motorcycle 420"],
        24: [0.500, 0.3125, 0.284, "Motorcycle 425"],
        25: [0.500, 0.335, 0.284, "Motorcycle 428"],
        26: [0.625, 0.400, 0.227, "Motorcycle 520"],
        27: [0.625, 0.400, 0.284, "Motorcycle 525"],
        28: [0.625, 0.400, 0.343, "Motorcycle 530"],
        29: [0.750, 0.400, 0.343, "Motorcycle 630"],
    }

    def __init__(self, obj):
        self.Type = "Sprocket"
        self.indx = 0
        self.sprockRef = []
        # As the UI file combobox has been translated the reference field needs populating here
        tempForm = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/PartDesign/SprocketFeature.ui"
        )
        for sKey in self.SprocketReferenceRollerTable:
            tempForm.comboBox_SprocketReference.setCurrentIndex(self.indx)
            self.sprockRef.append(tempForm.comboBox_SprocketReference.currentText())
            self.indx += 1
        self._ensure_properties(obj, is_restore=False)
        obj.addProperty(
            "App::PropertyEnumeration",
            "SprocketReference",
            "Sprocket",
            "Sprocket Reference",
            locked=True,
        )
        obj.SprocketReference = list(self.sprockRef)
        obj.Proxy = self

    def onDocumentRestored(self, obj):
        """hook used to migrate older versions of this object"""
        self._ensure_properties(obj, is_restore=True)

    def _ensure_properties(self, obj, is_restore):
        def ensure_property(type_, name, doc, default):
            if not hasattr(obj, name):
                obj.addProperty(type_, name, "Sprocket", locked=True)
                if callable(default):
                    setattr(obj, name, default())
                else:
                    setattr(obj, name, default)

        ensure_property("App::PropertyInteger", "NumberOfTeeth", "Number of gear teeth", 50)
        ensure_property("App::PropertyLength", "Pitch", "Chain Pitch", "0.375 in")
        ensure_property("App::PropertyLength", "RollerDiameter", "Roller Diameter", "0.20 in")
        ensure_property(
            "App::PropertyLength",
            "Thickness",
            "Thickness as stated in the reference specification",
            "0.11 in",
        )

    def execute(self, obj):
        w = fcsprocket.FCWireBuilder()
        sprocket.CreateSprocket(w, obj.Pitch.Value, obj.NumberOfTeeth, obj.RollerDiameter.Value)

        sprocketw = Part.Wire([o.toShape() for o in w.wire])
        obj.Shape = sprocketw
        obj.positionBySupport()
        return


class ViewProviderSprocket:
    """
    A View Provider for the Sprocket object
    """

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/PartDesign_Sprocket.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def setEdit(self, vobj, mode):
        taskd = SprocketTaskPanel(self.Object, mode)
        taskd.obj = vobj.Object
        taskd.update()
        FreeCADGui.Control.showDialog(taskd)
        return True

    def unsetEdit(self, vobj, mode):
        FreeCADGui.Control.closeDialog()
        return

    def dumps(self):
        return None

    def loads(self, state):
        return None


class SprocketTaskPanel:
    """
    The editmode TaskPanel for Sprocket objects
    """

    def __init__(self, obj, mode):
        self.obj = obj

        self.form = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/PartDesign/SprocketFeature.ui"
        )
        self.form.setWindowIcon(QtGui.QIcon(":/icons/PartDesign_Sprocket.svg"))

        QtCore.QObject.connect(
            self.form.Quantity_Pitch,
            QtCore.SIGNAL("valueChanged(double)"),
            self.pitchChanged,
        )
        QtCore.QObject.connect(
            self.form.Quantity_RollerDiameter,
            QtCore.SIGNAL("valueChanged(double)"),
            self.rollerDiameterChanged,
        )
        QtCore.QObject.connect(
            self.form.spinBox_NumberOfTeeth,
            QtCore.SIGNAL("valueChanged(int)"),
            self.numTeethChanged,
        )
        QtCore.QObject.connect(
            self.form.comboBox_SprocketReference,
            QtCore.SIGNAL("currentIndexChanged(int)"),
            self.sprocketReferenceChanged,
        )
        QtCore.QObject.connect(
            self.form.Quantity_Thickness,
            QtCore.SIGNAL("valueChanged(double)"),
            self.thicknessChanged,
        )

        self.update()

        if mode == 0:  # fresh created
            self.obj.Proxy.execute(self.obj)  # calculate once
            FreeCAD.Gui.SendMsgToActiveView("ViewFit")

    def transferTo(self):
        """
        Transfer from the dialog to the object
        """
        self.obj.NumberOfTeeth = self.form.spinBox_NumberOfTeeth.value()
        self.obj.Pitch = self.form.Quantity_Pitch.text()
        self.obj.RollerDiameter = self.form.Quantity_RollerDiameter.text()
        self.obj.SprocketReference = self.form.comboBox_SprocketReference.currentIndex()
        self.obj.Thickness = self.form.Quantity_Thickness.text()

    def transferFrom(self):
        """
        Transfer from the object to the dialog
        """
        self.form.spinBox_NumberOfTeeth.setValue(self.obj.NumberOfTeeth)
        self.form.Quantity_Pitch.setText(self.obj.Pitch.UserString)
        self.form.Quantity_RollerDiameter.setText(self.obj.RollerDiameter.UserString)
        self.form.comboBox_SprocketReference.setCurrentText(self.obj.SprocketReference)
        self.form.Quantity_Thickness.setText(self.obj.Thickness.UserString)

    def pitchChanged(self, value):
        self.obj.Pitch = value
        self.obj.Proxy.execute(self.obj)
        FreeCAD.Gui.SendMsgToActiveView("ViewFit")

    def sprocketReferenceChanged(self, size):
        self.obj.Pitch = str(Sprocket.SprocketReferenceRollerTable[size][0]) + " in"
        self.obj.RollerDiameter = str(Sprocket.SprocketReferenceRollerTable[size][1]) + " in"
        self.obj.Thickness = str(Sprocket.SprocketReferenceRollerTable[size][2]) + " in"
        self.obj.SprocketReference = self.obj.getEnumerationsOfProperty("SprocketReference")[size]
        self.form.Quantity_Pitch.setText(self.obj.Pitch.UserString)
        self.form.Quantity_RollerDiameter.setText(self.obj.RollerDiameter.UserString)
        self.form.Quantity_Thickness.setText(self.obj.Thickness.UserString)
        self.obj.Proxy.execute(self.obj)
        FreeCAD.Gui.SendMsgToActiveView("ViewFit")

    def rollerDiameterChanged(self, value):
        self.obj.RollerDiameter = value
        self.obj.Proxy.execute(self.obj)

    def numTeethChanged(self, value):
        self.obj.NumberOfTeeth = value
        self.obj.Proxy.execute(self.obj)
        FreeCAD.Gui.SendMsgToActiveView("ViewFit")

    def thicknessChanged(self, value):
        self.obj.Thickness = str(value)
        self.obj.Proxy.execute(self.obj)

    def getStandardButtons(self):
        return (
            QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel | QtGui.QDialogButtonBox.Apply
        )

    def clicked(self, button):
        if button == QtGui.QDialogButtonBox.Apply:
            self.transferTo()
            self.obj.Proxy.execute(self.obj)

    def update(self):
        self.transferFrom()

    def accept(self):
        self.transferTo()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()

    def reject(self):
        FreeCAD.ActiveDocument.removeObject(self.obj.Name)
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCAD.ActiveDocument.abortTransaction()



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2018 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM base constraint ViewProvider"
__author__ = "Markus Hovorka, Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package view_base_femconstraint
#  \ingroup FEM
#  \brief view provider for Python base constraint object

from FreeCAD import getResourceDir
from femviewprovider import view_base_femobject


class VPBaseFemConstraint(view_base_femobject.VPBaseFemObject):
    """Proxy View Provider for Pythons base constraint."""

    resource_symbol_dir = getResourceDir() + "Mod/Fem/Resources/symbols/"



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with edges."""
## @package edges
# \ingroup draftgeoutils
# \brief Provides various functions to work with edges.

import lazy_loader.lazy_loader as lz

import FreeCAD as App
import DraftVecUtils

from draftgeoutils.general import geomType, vec

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def findEdge(anEdge, aList):
    """Return True if edge is found in list of edges."""
    for e in range(len(aList)):
        if str(anEdge.Curve) == str(aList[e].Curve):
            if DraftVecUtils.equals(anEdge.Vertexes[0].Point, aList[e].Vertexes[0].Point):
                if DraftVecUtils.equals(anEdge.Vertexes[-1].Point, aList[e].Vertexes[-1].Point):
                    return e
    return None


def orientEdge(edge, normal=None, make_arc=False):
    """Re-orient the edge such that it is in the XY plane.

    Re-orients `edge` such that it is in the XY plane.
    If `normal` is passed, this is used as the basis for the rotation,
    otherwise the placement of `edge` is used.
    """
    # This 'normalizes' the placement to the xy plane
    edge = edge.copy()
    xyDir = App.Vector(0, 0, 1)
    base = App.Vector(0, 0, 0)

    if normal:
        angle = DraftVecUtils.angle(normal, xyDir) * App.Units.Radian
        axis = normal.cross(xyDir)
    else:
        axis = edge.Placement.Rotation.Axis
        angle = -1 * edge.Placement.Rotation.Angle * App.Units.Radian
    if axis == App.Vector(0.0, 0.0, 0.0):
        axis = App.Vector(0.0, 0.0, 1.0)
    if angle:
        edge.rotate(base, axis, angle)
    if isinstance(edge.Curve, Part.Line):
        return Part.LineSegment(edge.Curve, edge.FirstParameter, edge.LastParameter)
    elif make_arc and isinstance(edge.Curve, Part.Circle) and not edge.Closed:
        return Part.ArcOfCircle(
            edge.Curve, edge.FirstParameter, edge.LastParameter, edge.Curve.Axis.z > 0
        )
    elif make_arc and isinstance(edge.Curve, Part.Ellipse) and not edge.Closed:
        return Part.ArcOfEllipse(
            edge.Curve, edge.FirstParameter, edge.LastParameter, edge.Curve.Axis.z > 0
        )
    return edge.Curve


def isSameLine(e1, e2):
    """Return True if the 2 edges are lines and have the same points."""
    if not isinstance(e1.Curve, Part.LineSegment):
        return False
    if not isinstance(e2.Curve, Part.LineSegment):
        return False

    if DraftVecUtils.equals(e1.Vertexes[0].Point, e2.Vertexes[0].Point) and DraftVecUtils.equals(
        e1.Vertexes[-1].Point, e2.Vertexes[-1].Point
    ):
        return True
    elif DraftVecUtils.equals(e1.Vertexes[-1].Point, e2.Vertexes[0].Point) and DraftVecUtils.equals(
        e1.Vertexes[0].Point, e2.Vertexes[-1].Point
    ):
        return True
    return False


def is_line(bspline):
    """Return True if the given BSpline curve is a straight line."""

    # previous implementation may fail for a multipole straight spline due
    # a second order error in tolerance, which introduce a difference of 1e-14
    # in the values of the tangents. Also, may fail on a periodic spline.
    #    step = bspline.LastParameter/10
    #    b = bspline.tangent(0)
    #
    #    for i in range(10):
    #        if bspline.tangent(i * step) != b:
    #            return False

    start_point = bspline.StartPoint
    end_point = bspline.EndPoint
    dist_start_end = end_point.distanceToPoint(start_point)
    if abs(bspline.length() - dist_start_end) < 1e-7:
        return True

    return False


def invert(shape):
    """Return an inverted copy of the edge or wire contained in the shape."""
    if shape.ShapeType == "Wire":
        edges = [invert(edge) for edge in shape.OrderedEdges]
        edges.reverse()
        return Part.Wire(edges)
    elif shape.ShapeType == "Edge":
        if len(shape.Vertexes) == 1:
            return shape
        if geomType(shape) == "Line":
            return Part.LineSegment(shape.Vertexes[-1].Point, shape.Vertexes[0].Point).toShape()
        elif geomType(shape) == "Circle":
            mp = findMidpoint(shape)
            return Part.Arc(shape.Vertexes[-1].Point, mp, shape.Vertexes[0].Point).toShape()
        elif geomType(shape) in ["BSplineCurve", "BezierCurve"]:
            if isLine(shape.Curve):
                return Part.LineSegment(shape.Vertexes[-1].Point, shape.Vertexes[0].Point).toShape()

        print("DraftGeomUtils.invert: unable to invert", shape.Curve)
        return shape
    else:
        print("DraftGeomUtils.invert: unable to handle", shape.ShapeType)
        return shape


def findMidpoint(edge):
    """Return the midpoint of an edge."""
    if edge.Length == 0:
        return None
    else:
        return edge.valueAt(edge.Curve.parameterAtDistance(edge.Length / 2, edge.FirstParameter))


def getTangent(edge, from_point=None):
    """Return the tangent to an edge, including BSpline and circular arcs.

    If from_point is given, it is used to calculate the tangent,
    only useful for a circular arc.
    """
    if geomType(edge) == "Line":
        return vec(edge)

    elif geomType(edge) == "BSplineCurve" or geomType(edge) == "BezierCurve":
        if not from_point:
            return None
        cp = edge.Curve.parameter(from_point)
        return edge.Curve.tangent(cp)[0]

    elif geomType(edge) == "Circle":
        if not from_point:
            v1 = edge.Vertexes[0].Point.sub(edge.Curve.Center)
        else:
            v1 = from_point.sub(edge.Curve.Center)
        return v1.cross(edge.Curve.Axis)

    return None


def get_referenced_edges(property_value):
    """Return the Edges referenced by the value of a App:PropertyLink, App::PropertyLinkList,
    App::PropertyLinkSub or App::PropertyLinkSubList property."""
    edges = []
    if not isinstance(property_value, list):
        property_value = [property_value]
    for element in property_value:
        if hasattr(element, "Shape") and element.Shape:
            edges += element.Shape.Edges
        elif isinstance(element, tuple) and len(element) == 2:
            object, subelement_names = element
            if hasattr(object, "Shape") and object.Shape:
                if len(subelement_names) == 1 and subelement_names[0] == "":
                    edges += object.Shape.Edges
                else:
                    for subelement_name in subelement_names:
                        if subelement_name.startswith("Edge"):
                            edge_number = int(subelement_name.lstrip("Edge")) - 1
                            if edge_number < len(object.Shape.Edges):
                                edges.append(object.Shape.Edges[edge_number])
    return edges


# compatibility layer

isLine = is_line

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import importlib

__title__ = "CAM Icon ViewProvider"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "ViewProvider who's main and only task is to assign an icon."

translate = FreeCAD.Qt.translate

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class ViewProvider(object):
    """Generic view provider to assign an icon."""

    def __init__(self, vobj, icon):
        self.icon = icon
        self.attach(vobj)

        self.editModule = None
        self.editCallback = None

        vobj.Proxy = self

    def attach(self, vobj):
        self.vobj = vobj
        self.obj = vobj.Object

    def dumps(self):
        attrs = {"icon": self.icon}
        if hasattr(self, "editModule"):
            attrs["editModule"] = self.editModule
            attrs["editCallback"] = self.editCallback
        return attrs

    def loads(self, state):
        self.icon = state["icon"]
        if state.get("editModule", None):
            self.editModule = state["editModule"]
            self.editCallback = state["editCallback"]

    def getIcon(self):
        return ":/icons/CAM_{}.svg".format(self.icon)

    def onEdit(self, callback):
        self.editModule = callback.__module__
        self.editCallback = callback.__name__

    def _onEditCallback(self, edit):
        if hasattr(self, "editModule"):
            mod = importlib.import_module(self.editModule)
            callback = getattr(mod, self.editCallback)
            callback(self.obj, self.vobj, edit)

    def setEdit(self, vobj=None, mode=0):
        if 0 == mode:
            self._onEditCallback(True)
        elif 1 == mode:
            FreeCADGui.runCommand("Std_TransformManip")
            return True
        return False

    def unsetEdit(self, vobj, mode):
        if 0 == mode:
            self._onEditCallback(False)

    def setupContextMenu(self, vobj, menu):
        Path.Log.track()
        from PySide import QtGui

        edit = translate("Path", "Edit")
        action = QtGui.QAction(edit, menu)
        action.triggered.connect(self._editInContextMenuTriggered)
        menu.addAction(action)

    def _editInContextMenuTriggered(self, checked):
        self.setEdit()


_factory = {}


def Attach(vobj, name):
    """Attach(vobj, name) ... attach the appropriate view provider to the view object.
    If no view provider was registered for the given name a default IconViewProvider is created."""

    Path.Log.track(vobj.Object.Label, name)
    global _factory
    for key, value in _factory.items():
        if key == name:
            return value(vobj, name)
    Path.Log.track(vobj.Object.Label, name, "PathIconViewProvider")
    return ViewProvider(vobj, name)


def RegisterViewProvider(name, provider):
    """RegisterViewProvider(name, provider) ... if an IconViewProvider is created for an object with the given name
    an instance of provider is used instead."""

    Path.Log.track(name)
    global _factory
    _factory[name] = provider



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import pathlib
import FreeCAD
import Path
import os
from typing import Dict, List, Any, Mapping, Optional, Tuple, Type, cast
import zipfile
import xml.etree.ElementTree as ET
import io
import tempfile
from ...assets import Asset, AssetUri, AssetSerializer, DummyAssetSerializer
from ...camassets import cam_assets
from ..doc import (
    find_shape_object,
    get_object_properties,
    get_unset_value_for,
    update_shape_object_properties,
    ShapeDocFromBytes,
)
from .icon import ToolBitShapeIcon


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class ToolBitShape(Asset):
    """Abstract base class for tool bit shapes."""

    asset_type: str = "toolbitshape"

    # The name is used...
    #   1. as a base for the default filename. E.g. if the name is
    #      "Endmill", then by default the file is "endmill.fcstd".
    #   2. to identify the shape class from a shape.fcstd file.
    #      Upon loading a shape, the name of the body in the shape
    #      file is read. It MUST match one of the names.
    name: str

    # Aliases exist for backward compatibility. If an existing .fctb file
    # references a shape such as "v-bit.fctb", and that shape file cannot
    # be found, then we can attempt to find a shape class from the string
    # "v-bit", "vbit", etc.
    aliases: Tuple[str, ...] = tuple()

    def __init__(self, id: str, **kwargs: Any):
        """
        Initialize the shape.

        Args:
            id (str): The unique identifier for the shape.
            **kwargs: Keyword arguments for shape parameters (e.g., Diameter).
                      Values should be FreeCAD.Units.Quantity where applicable.
        """
        # _params will be populated with default values after loading
        self._params: Dict[str, Any] = {}

        # Stores default parameter values loaded from the FCStd file
        self._defaults: Dict[str, Any] = {}

        # Stores the FreeCAD property types for each parameter
        self._param_types: Dict[str, str] = {}

        # Keeps the loaded FreeCAD document content for this instance
        self._data: Optional[bytes] = None

        self.id: str = id

        self.icon: Optional[ToolBitShapeIcon] = None

        # Assign parameters
        for param, value in kwargs.items():
            self.set_parameter(param, value)

    def __str__(self):
        params_str = ", ".join(f"{name}={val}" for name, val in self._params.items())
        return f"{self.name}({params_str})"

    def __repr__(self):
        return self.__str__()

    def get_id(self) -> str:
        """
        Get the ID of the shape.

        Returns:
            str: The ID of the shape.
        """
        return self.id

    @property
    def is_builtin(self) -> bool:
        """
        Check if this shape is a built-in shape.

        Returns:
            bool: True if the shape is built-in, False otherwise.
        """
        return self.name.lower() == self.id

    @classmethod
    def _get_shape_class_from_doc(cls, doc: "FreeCAD.Document") -> Type["ToolBitShape"]:
        # Find the Body object to identify the shape type
        body_obj = find_shape_object(doc)
        if not body_obj:
            raise ValueError(f"No 'PartDesign::Body' object found in {doc}")

        # Find the correct subclass based on the body label
        shape_classes = {c.name: c for c in ToolBitShape.__subclasses__()}
        shape_class = shape_classes.get(body_obj.Label)
        if not shape_class:
            custom = ToolBitShape.get_subclass_by_name("Custom")
            assert custom is not None, "BUG: Custom tool class not found"
            return custom
        return shape_class

    @classmethod
    def get_shape_class_from_id(
        cls,
        shape_id: str,
        shape_type: Optional[str] = None,
        default: Optional[Type["ToolBitShape"]] = None,
    ) -> Optional[Type["ToolBitShape"]]:
        """
        Extracts the shape class from the given ID and shape_type, retrieving it
        from the asset manager if necessary.
        """
        # Best method: if the shape-type is specified, use that.
        if shape_type:
            return ToolBitShape.get_subclass_by_name(shape_type)

        # If no shape type is specified, try to find the shape class from the ID.
        shape_class = ToolBitShape.get_subclass_by_name(shape_id)
        if shape_class:
            return shape_class

        # If that also fails, try to load the shape to get the class.
        Path.Log.debug(
            f'Failed to infer shape type from "{shape_id}", trying to load'
            f' the shape "{shape_id}" to determine the class. This may'
            " negatively impact performance."
        )
        shape_asset_uri = ToolBitShape.resolve_name(shape_id)
        try:
            data = cam_assets.get_raw(shape_asset_uri)
        except FileNotFoundError:
            pass  # rely on fallback below
        else:
            try:
                shape_class = ToolBitShape.get_shape_class_from_bytes(data)
            except ValueError:
                pass
            else:
                return shape_class

        # Otherwise use the default, if we have one.
        shape_types = [c.name for c in ToolBitShape.__subclasses__()]
        if default is not None:
            Path.Log.debug(
                f'Failed to infer shape type from {shape_id}, using "{default.name}".'
                f" To fix, name the body in the shape file to one of: {shape_types}"
            )
            return default

        # Default to Custom if nothing else works
        return ToolBitShape.get_subclass_by_name("Custom")

    @classmethod
    def get_shape_class_from_bytes(cls, data: bytes) -> Type["ToolBitShape"]:
        """
        Identifies the ToolBitShape subclass from the raw bytes of an FCStd file
        by parsing the XML content to find the Body label.

        Args:
            data (bytes): The raw bytes of the .FCStd file.

        Returns:
            Type[ToolBitShape]: The appropriate ToolBitShape subclass.

        Raises:
            ValueError: If the data is not a valid FCStd file, Document.xml is
                        missing, no Body object is found, or the Body label
                        does not match a known shape name.
        """
        try:
            # FCStd files are zip archives
            with zipfile.ZipFile(io.BytesIO(data)) as zf:
                # Read Document.xml from the archive
                with zf.open("Document.xml") as doc_xml_file:
                    tree = ET.parse(doc_xml_file)
                    root = tree.getroot()

            # Extract name of the main Body from XML tree using xpath.
            # The body should be a PartDesign::Body, and its label is
            # stored in an Property element with a matching name.
            body_label = None
            xpath = './/Object[@name="Body"]//Property[@name="Label"]/String'
            body_label_elem = root.find(xpath)
            if body_label_elem is not None:
                body_label = body_label_elem.get("value")

            if not body_label:
                raise ValueError(
                    "No 'Label' property found for 'PartDesign::Body' object using XPath"
                )

            # Find the correct subclass based on the body label
            shape_class = cls.get_subclass_by_name(body_label)
            if shape_class:
                return shape_class

            # All else fails, treat the shape as a custom shape.
            custom = ToolBitShape.get_subclass_by_name("Custom")
            assert custom is not None, "BUG: Custom tool class not found"
            return custom

        except zipfile.BadZipFile:
            raise ValueError("Invalid FCStd file data (not a valid zip archive)")
        except KeyError:
            raise ValueError("Invalid FCStd file data (Document.xml not found)")
        except ET.ParseError:
            raise ValueError("Error parsing Document.xml")
        except Exception as e:
            # Catch any other unexpected errors during parsing
            raise ValueError(f"Error processing FCStd data: {e}")

    @classmethod
    def _find_property_object(cls, doc: "FreeCAD.Document") -> Optional["FreeCAD.DocumentObject"]:
        """
        Find the PropertyBag object named "Attributes" in a document.

        Args:
            doc (FreeCAD.Document): The document to search within.

        Returns:
            Optional[FreeCAD.DocumentObject]: The found object or None.
        """
        for o in doc.Objects:
            # Check if the object has a Label property and if its value is "Attributes"
            # This seems to be the convention in the shape files.
            if hasattr(o, "Label") and o.Label == "Attributes":
                # We assume this object holds the parameters.
                # Further type checking (e.g., for App::FeaturePython or PropertyBag)
                # could be added if needed, but Label check might be sufficient.
                return o
        return None

    @classmethod
    def extract_dependencies(cls, data: bytes, serializer: Type[AssetSerializer]) -> List[AssetUri]:
        """
        Extracts URIs of dependencies from the raw bytes of an FCStd file.
        For ToolBitShape, this is the associated ToolBitShapeIcon, identified
        by the same ID as the shape asset.
        """
        Path.Log.debug(f"ToolBitShape.extract_dependencies called for {cls.__name__}")
        assert (
            serializer == DummyAssetSerializer
        ), f"ToolBitShape supports only native import, not {serializer}"

        # A ToolBitShape asset depends on a ToolBitShapeIcon asset with the same ID.
        # We need to extract the shape ID from the FCStd data.
        try:
            # Open the shape data temporarily to get the Body label, which can
            # be used to derive the ID if needed, or assume the ID is available
            # in the data somehow (e.g., in a property).
            # For now, let's assume the ID is implicitly the asset name derived
            # from the Body label.
            shape_class = cls.get_shape_class_from_bytes(data)
            shape_id = shape_class.name.lower()  # Assuming ID is lowercase name

            # Construct the URI for the corresponding icon asset
            svg_uri = AssetUri.build(
                asset_type="toolbitshapesvg",
                asset_id=shape_id + ".svg",
            )
            png_uri = AssetUri.build(
                asset_type="toolbitshapepng",
                asset_id=shape_id + ".png",
            )
            return [svg_uri, png_uri]

        except Exception as e:
            # If we can't extract the shape ID or something goes wrong,
            # assume no dependencies for now.
            Path.Log.error(f"Failed to extract dependencies from shape data: {e}")
            return []

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
        serializer: Type[AssetSerializer],
    ) -> "ToolBitShape":
        """
        Create a ToolBitShape instance from the raw bytes of an FCStd file.

        Identifies the correct subclass based on the Body label in the file,
        loads parameters, and caches the document content.

        Args:
            data (bytes): The raw bytes of the .FCStd file.
            id (str): The unique identifier for the shape.
            dependencies (Optional[Mapping[AssetUri, Any]]): A mapping of
                  resolved dependencies. If None, shallow load was attempted.

        Returns:
            ToolBitShape: An instance of the appropriate ToolBitShape subclass.

        Raises:
            ValueError: If the data cannot be opened, no Body or PropertyBag
                        is found, or the Body label does not match a known
                        shape name.
            Exception: For other potential FreeCAD errors during loading.
        """
        assert serializer == DummyAssetSerializer, "ToolBitShape supports only native import"
        Path.Log.debug(f"{id}: ToolBitShape.from_bytes called with {len(data)} bytes")

        # Open the shape data temporarily to get the Body label and parameters
        with ShapeDocFromBytes(data) as temp_doc:
            if not temp_doc:
                # This case might be covered by ShapeDocFromBytes exceptions,
                # but keeping for clarity.
                raise ValueError("Failed to open shape document from bytes")

            # Determine the specific subclass of ToolBitShape.
            try:
                shape_class = ToolBitShape.get_shape_class_from_bytes(data)
            except Exception as e:
                Path.Log.debug(f"{id}: Failed to determine shape class from bytes: {e}")
                shape_class = ToolBitShape.get_shape_class_from_id("Custom")
            if shape_class is None:
                # This should ideally not happen due to get_shape_class_from_bytes fallback
                # but added for linter satisfaction.
                raise ValueError("Shape class could not be determined.")

            # Load properties from the temporary document
            props_obj = ToolBitShape._find_property_object(temp_doc)
            if not props_obj:
                raise ValueError("No 'Attributes' PropertyBag object found in document bytes")

            # loaded_raw_params will now be Dict[str, Tuple[Any, str]]
            loaded_raw_params = get_object_properties(props_obj, exclude_groups=["", "Base"])

            # Separate values and types, and populate _param_types
            loaded_params = {}
            loaded_param_types = {}
            for name, (value, type_id) in loaded_raw_params.items():
                loaded_params[name] = value
                loaded_param_types[name] = type_id

            # For now, we log missing parameters, but do not raise an error.
            # This allows for more flexible shape files that may not have all
            # parameters set, while still warning the user.
            # In the future, we may want to raise an error if critical parameters
            # are missing.
            expected_params = shape_class.get_expected_shape_parameters()
            missing_params = [
                name
                for name in expected_params
                if name not in loaded_params or loaded_params[name] is None
            ]
            if missing_params:
                Path.Log.error(
                    f"Validation error: Object '{props_obj.Label}' in document {id} "
                    f"is missing parameters for {shape_class.__name__}: {', '.join(missing_params)}."
                    f" In future releases, these shapes will not load!"
                )
                for param in missing_params:
                    param_type = shape_class.get_schema_property_type(param)
                    loaded_params[param] = get_unset_value_for(param_type)
                    loaded_param_types[param] = param_type  # Store the type for missing params

            # Instantiate the specific subclass with the provided ID
            instance = shape_class(id=id)
            instance._data = data  # Keep the byte content
            instance._defaults = loaded_params
            instance._param_types = loaded_param_types
            Path.Log.debug(f"Params: {instance._params} {instance._defaults}")
            instance._params = instance._defaults | instance._params

            if dependencies:  # dependencies is None = shallow load
                # Assign resolved dependencies (like the icon) to the instance
                # The icon has the same ID as the shape, with .png or .svg appended.
                icon_uri = AssetUri.build(
                    asset_type="toolbitshapesvg",
                    asset_id=id + ".svg",
                )
                instance.icon = cast(ToolBitShapeIcon, dependencies.get(icon_uri))
                if not instance.icon:
                    icon_uri = AssetUri.build(
                        asset_type="toolbitshapepng",
                        asset_id=id + ".png",
                    )
                    instance.icon = cast(ToolBitShapeIcon, dependencies.get(icon_uri))

            return instance

    def to_bytes(self, serializer: Type[AssetSerializer]) -> bytes:
        """
        Serializes a ToolBitShape object to bytes (e.g., an fcstd file).
        This is required by the Asset interface.
        """
        assert serializer == DummyAssetSerializer, "ToolBitShape supports only native export"
        doc = None
        try:
            # Create a new temporary document
            doc = FreeCAD.newDocument("TemporaryShapeDoc", hidden=True)

            # Add the shape's body to the temporary document
            self.make_body(doc)

            # Recompute the document to ensure the body is created
            doc.recompute()

            # Save the temporary document to a temporary file
            # We cannot use NamedTemporaryFile on Windows, because there
            # doc.saveAs() may not have permission to access the tempfile
            # while the NamedTemporaryFile is open.
            # So we use TemporaryDirectory instead, to ensure cleanup while
            # still having a the temporary file inside it.
            with tempfile.TemporaryDirectory() as thedir:
                temp_file_path = pathlib.Path(thedir, "temp.FCStd")
                doc.saveAs(str(temp_file_path))
                return temp_file_path.read_bytes()

        finally:
            # Clean up the temporary document
            if doc:
                FreeCAD.closeDocument(doc.Name)

    @classmethod
    def from_file(cls, filepath: pathlib.Path, **kwargs: Any) -> "ToolBitShape":
        """
        Create a ToolBitShape instance from an FCStd file.

        Reads the file bytes and delegates to from_bytes().

        Args:
            filepath (pathlib.Path): Path to the .FCStd file.
            **kwargs: Keyword arguments for shape parameters to override defaults.

        Returns:
            ToolBitShape: An instance of the appropriate ToolBitShape subclass.

        Raises:
            FileNotFoundError: If the file does not exist.
            ValueError: If the file cannot be opened, no Body or PropertyBag
                        is found, or the Body label does not match a known
                        shape name.
            Exception: For other potential FreeCAD errors during loading.
        """
        if not filepath.exists():
            raise FileNotFoundError(f"Shape file not found: {filepath}")
        Path.Log.debug(f"{id}: ToolBitShape.from_file called with {filepath}")

        try:
            data = filepath.read_bytes()
            # Extract the ID from the filename (without extension)
            shape_id = filepath.stem
            # Pass an empty dictionary for dependencies when loading from a single file
            # TODO: pass ToolBitShapeIcon as a dependency
            instance = cls.from_bytes(data, shape_id, {}, DummyAssetSerializer)
            # Apply kwargs parameters after loading from bytes
            if kwargs:
                instance.set_parameters(**kwargs)
            return instance
        except (FileNotFoundError, ValueError) as e:
            raise e
        except Exception as e:
            raise RuntimeError(f"Failed to create shape from {filepath}: {e}")

    @classmethod
    def get_subclass_by_name(
        cls, name: str, default: Optional[Type["ToolBitShape"]] = None
    ) -> Optional[Type["ToolBitShape"]]:
        """
        Retrieves a ToolBitShape class by its name or alias.
        """
        name = name.lower()
        for thecls in cls.__subclasses__():
            if (
                thecls.name.lower() == name
                or thecls.__name__.lower() == name
                or name in thecls.aliases
            ):
                return thecls
        return default

    @classmethod
    def guess_subclass_from_name(
        cls, name: str, default: Optional[Type["ToolBitShape"]] = None
    ) -> Optional[Type["ToolBitShape"]]:
        """
        Retrieves a ToolBitShape class by its name or alias.
        """
        name = name.lower()
        for thecls in cls.__subclasses__():
            if thecls.name.lower() in name or thecls.__name__.lower() in name:
                return thecls
            for alias in thecls.aliases:
                if alias.lower() in name:
                    return thecls
        return default

    @classmethod
    def resolve_name(cls, identifier: str) -> AssetUri:
        """
        Resolves an identifier (name, filename, or URI) to a Uri object.
        """
        # 1. If the input is a url string, return the AssetUri for it.
        if AssetUri.is_uri(identifier):
            return AssetUri(identifier)

        # 2. If the input is a filename (with extension), assume the asset
        #    name is the base name.
        asset_name = identifier
        if pathlib.Path(identifier).suffix.lower() == ".fcstd":
            asset_name = os.path.splitext(os.path.basename(identifier))[0]

        # 3. Construct the Uri using AssetUri.build() and return it
        return AssetUri.build(
            asset_type="toolbitshape",
            asset_id=asset_name,
        )

    @classmethod
    def schema(cls) -> Mapping[str, Tuple[str, str]]:
        """
        Subclasses must define the dictionary mapping parameter names to
        translations and FreeCAD property type strings (e.g.,
        'App::PropertyLength').

        The schema defines any parameters that MUST be in the shape file.
        Any attempt to load a shape file that does not match the schema
        will cause an error.
        """
        raise NotImplementedError

    @property
    def label(self) -> str:
        """Return a user friendly, translatable display name."""
        raise NotImplementedError

    def reset_parameters(self):
        """Reset parameters to their default values."""
        self._params.update(self._defaults)

    def get_parameter_label(self, param_name: str) -> str:
        """
        Get the user-facing label for a given parameter name.
        """
        str_param_name = str(param_name)
        entry = self.schema().get(param_name)
        return entry[0] if entry else str_param_name

    @classmethod
    def get_schema_property_type(cls, param_name: str) -> str:
        """
        Get the FreeCAD property type string for a given parameter name.
        """
        return cls.schema()[param_name][1]

    def get_parameter_property_type(
        self, param_name: str, default: str = "App::PropertyString"
    ) -> str:
        """
        Get the FreeCAD property type string for a given parameter name.
        """
        try:
            return self.get_schema_property_type(param_name)
        except KeyError:
            try:
                return self._param_types[param_name]
            except KeyError:
                return default

    def _normalize_value(self, name: str, value: Any) -> Any:
        """
        Normalize the value for a parameter based on its expected type.
        This is a placeholder for any type-specific normalization logic.

        Args:
            name (str): The name of the parameter.
            value: The value to normalize.

        Returns:
            The normalized value, potentially converted to a FreeCAD.Units.Quantity.
        """
        prop_type = self.get_parameter_property_type(name)
        if prop_type in ("App::PropertyDistance", "App::PropertyLength", "App::PropertyAngle"):
            return FreeCAD.Units.Quantity(value)
        elif prop_type == "App::PropertyInteger":
            return int(value)
        elif prop_type == "App::PropertyFloat":
            return float(value)
        elif prop_type == "App::PropertyBool":
            if value in ("True", "true", "1"):
                return True
            elif value in ("False", "false", "0"):
                return False
            return bool(value)
        return str(value)

    def get_parameters(self) -> Dict[str, Any]:
        """
        Get the dictionary of current parameters and their values.

        Returns:
            dict: A dictionary mapping parameter names to their values.
        """
        return {name: self._normalize_value(name, value) for name, value in self._params.items()}

    def get_parameter(self, name: str) -> Any:
        """
        Get the value of a specific parameter.

        Args:
            name (str): The name of the parameter.

        Returns:
            The value of the parameter (often a FreeCAD.Units.Quantity).

        Raises:
            KeyError: If the parameter name is not valid for this shape.
        """
        if name not in self.schema():
            raise KeyError(f"Shape '{self.name}' has no parameter '{name}'")
        return self._normalize_value(name, self._params[name])

    def set_parameter(self, name: str, value: Any):
        """
        Set the value of a specific parameter.

        Args:
            name (str): The name of the parameter.
            value: The new value for the parameter. Should be compatible
                   with the expected type (e.g., FreeCAD.Units.Quantity).

        Raises:
            KeyError: If the parameter name is not valid for this shape.
        """
        self._params[name] = self._normalize_value(name, value)

    def set_parameters(self, **kwargs):
        """
        Set multiple parameters using keyword arguments.

        Args:
            **kwargs: Keyword arguments where keys are parameter names.
        """
        for name, value in kwargs.items():
            try:
                self.set_parameter(name, value)
            except KeyError:
                Path.Log.debug(f"Ignoring unknown parameter '{name}' for shape '{self.name}'.\n")

    @classmethod
    def get_expected_shape_parameters(cls) -> List[str]:
        """
        Get a list of parameter names expected by this shape class based on
        its schema.

        Returns:
            list[str]: List of parameter names.
        """
        return list(cls.schema().keys())

    def make_body(self, doc: "FreeCAD.Document"):
        """
        Generates the body of the ToolBitShape and copies it to the provided
        document.
        """
        assert self._data is not None
        with ShapeDocFromBytes(self._data) as tmp_doc:
            shape = find_shape_object(tmp_doc)
            if not shape:
                FreeCAD.Console.PrintWarning(
                    "No suitable shape object found in document. " "Cannot create solid shape.\n"
                )
                return None

            props = self._find_property_object(tmp_doc)
            if not props:
                FreeCAD.Console.PrintWarning(
                    "No suitable shape object found in document. " "Cannot create solid shape.\n"
                )
                return None

            update_shape_object_properties(props, self.get_parameters())

            # Recompute the document to apply property changes
            tmp_doc.recompute()

            # Temporarily disable duplicate labels to let FreeCAD automatically
            # make labels unique during the copy operation

            param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Document")
            original_setting = param.GetBool("DuplicateLabels", False)

            try:
                # Disable duplicate labels temporarily
                param.SetBool("DuplicateLabels", False)

                # Copy the body - FreeCAD will now automatically make labels unique
                copied_shape = doc.copyObject(shape, True)

                return copied_shape
            finally:
                # Restore the original setting
                param.SetBool("DuplicateLabels", original_setting)

        """
        Retrieves the thumbnail data for the tool bit shape in PNG format.
        """

    def get_parameter_type(self, name: str) -> str:
        """
        Get the FreeCAD property type string for a given parameter name,
        as loaded from the FCStd file.
        """
        return self._param_types.get(name, "App::PropertyString")

    def get_icon(self) -> Optional[ToolBitShapeIcon]:
        """
        Get the associated ToolBitShapeIcon instance. Tries to load one from
        the asset manager if none was assigned.

        Returns:
            Optional[ToolBitShapeIcon]: The icon instance, or None if none found.
        """
        if self.icon:
            return self.icon

        # Try to get a matching SVG from the asset manager.
        self.icon = cast(
            ToolBitShapeIcon, cam_assets.get_or_none(f"toolbitshapesvg://{self.id}.svg")
        )
        if self.icon:
            return self.icon

        # Try to get a matching PNG from the asset manager.
        self.icon = cast(
            ToolBitShapeIcon, cam_assets.get_or_none(f"toolbitshapepng://{self.id}.png")
        )
        if self.icon:
            return self.icon
        return None

    def get_thumbnail(self) -> Optional[bytes]:
        """
        Retrieves the thumbnail data for the tool bit shape in PNG format,
        as embedded in the shape file.
        """
        if not self._data:
            return None
        with zipfile.ZipFile(io.BytesIO(self._data)) as zf:
            try:
                with zf.open("thumbnails/Thumbnail.png", "r") as tn:
                    return tn.read()
            except KeyError:
                pass
        return None



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Syres                                              *
# *   Copyright (c) 2024 Furgo                                              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provide the grid observer for the Draft and BIM workbenches."""

import FreeCAD


# View observer code to update the Draft_ToggleGrid command button to reflect
# the grid's visibility status.
# Based on view observer code to update the Draft Tray.
if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtWidgets
    from draftutils import gui_utils
    from draftutils.todo import ToDo

    def _update_grid_gui():
        """Update function executed by the callback to refresh the Toggle Grid
        button state on all toolbars and menus.
        """

        try:
            # Get the active view.
            view = gui_utils.get_3d_view()

            # If the active view is not a 3D view, uncheck the button.
            if view is None:
                _set_grid_button_state(False)
                return

            # Otherwise, if there is a view, update the button's checked state.

            # Retrieve the associated grid for each MDI document.
            # [1] is the index where the grid is stored in the
            # trackers list. See setTrackers().
            view_idx = FreeCADGui.Snapper.trackers[0].index(view)
            grid = FreeCADGui.Snapper.trackers[1][view_idx]
            if grid.Visible:
                _set_grid_button_state(True)
            else:
                _set_grid_button_state(False)

        except Exception:
            pass

    def _view_observer_callback():
        """Callback function to update the Toggle Grid button.

        The update will only happen if either the Draft or BIM
        workbenches are active.
        """

        ToDo.delay(_update_grid_gui, None)

    _view_observer_active = False

    def _view_observer_setup():
        """Start or stop the grid observer.

        Start: it is intended to happen when either the Draft or BIM workbench
        are activated. The _view_observer_callback is connected to Qt's
        subWindowActivated signal. The MDI area emits the subWindowActivated()
        signal when the active window changes.

        Stop: This happens when either the Draft or BIM workbenches are
        deactivated and is the reverse of the start operation.
        """

        mw = FreeCADGui.getMainWindow()
        mdi = mw.findChild(QtWidgets.QMdiArea)
        global _view_observer_active
        if not _view_observer_active:
            # Connect callback to subWindowActivated signal
            mdi.subWindowActivated.connect(_view_observer_callback)
            _view_observer_active = True
            # Trigger initial grid button update
            _view_observer_callback()
        else:
            mdi.subWindowActivated.disconnect(_view_observer_callback)
            _view_observer_active = False

    def _set_grid_button_state(button_check):
        """Sets the checked state of the Draft_ToggleGrid command.
        This is then reflected on the associated "Toggle Grid" buttons of every
        toolbar and menu.

        Args:
            button_check (bool): if True, check the grid button.

        Returns:
            None
        """

        action = FreeCADGui.Command.get("Draft_ToggleGrid").getAction()[0]
        action.setCheckable(True)
        action.setChecked(button_check)
